from typing import Optional
import pandas as pd
from uuid import UUID
from sempy_labs._helper_functions import (
    resolve_workspace_name_and_id,
    _base_api,
    resolve_item_id,
    _create_dataframe,
)


def get_report_datasources(
    report: str | UUID,
    workspace: Optional[str | UUID] = None,
) -> pd.DataFrame:
    """
    Returns a list of data sources for the specified paginated report (RDL) from the specified workspace.

    Parameters
    ----------
    report : str | uuid.UUID
        Name or ID of the Power BI report.
    workspace : str | uuid.UUID, default=None
        The name or ID of the Fabric workspace in which the report resides.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of data sources for the specified paginated report (RDL) from the specified workspace.
    """

    columns = {
        "Report Name": "str",
        "Report Id": "str",
        "Datasource Id": "str",
        "Datasource Type": "str",
        "Gateway Id": "str",
        "Server": "str",
        "Database": "str",
    }
    df = _create_dataframe(columns=columns)

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    report_id = resolve_item_id(
        item=report, type="PaginatedReport", workspace=workspace
    )

    response = _base_api(
        request=f"v1.0/myorg/groups/{workspace_id}/reports/{report_id}/datasources"
    )

    for i in response.json().get("value", []):
        conn = i.get("connectionDetails", {})
        new_data = {
            "Report Name": report,
            "Report Id": report_id,
            "Datasource Id": i.get("datasourceId"),
            "Datasource Type": i.get("datasourceType"),
            "Gateway Id": i.get("gatewayId"),
            "Server": conn.get("server") if conn else None,
            "Database": conn.get("database") if conn else None,
        }

        df = pd.concat([df, pd.DataFrame(new_data, index=[0])], ignore_index=True)

    return df



import pandas as pd
from typing import Optional
from sempy_labs._helper_functions import (
    resolve_workspace_name_and_id,
    _base_api,
    resolve_item_id,
    _create_dataframe,
    _conv_b64,
    _decode_b64,
    delete_item,
    create_item,
    get_item_definition,
)
from uuid import UUID
import sempy_labs._icons as icons


def create_eventhouse(
    name: str,
    definition: Optional[dict],
    description: Optional[str] = None,
    workspace: Optional[str | UUID] = None,
):
    """
    Creates a Fabric eventhouse.

    This is a wrapper function for the following API: `Items - Create Eventhouse <https://learn.microsoft.com/rest/api/fabric/environment/items/create-eventhouse>`_.

    Parameters
    ----------
    name: str
        Name of the eventhouse.
    definition : dict
        The definition (EventhouseProperties.json) of the eventhouse.
    description : str, default=None
        A description of the environment.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    if definition is not None and not isinstance(definition, dict):
        raise ValueError(f"{icons.red_dot} The definition must be a dictionary.")

    definition_payload = (
        {
            "parts": [
                {
                    "path": "EventhouseProperties.json",
                    "payload": _conv_b64(definition),
                    "payloadType": "InlineBase64",
                }
            ]
        }
        if definition is not None
        else None
    )

    create_item(
        name=name,
        type="Eventhouse",
        workspace=workspace,
        description=description,
        definition=definition_payload,
    )


def list_eventhouses(workspace: Optional[str | UUID] = None) -> pd.DataFrame:
    """
    Shows the eventhouses within a workspace.

    This is a wrapper function for the following API: `Items - List Eventhouses <https://learn.microsoft.com/rest/api/fabric/environment/items/list-eventhouses>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the eventhouses within a workspace.
    """

    columns = {
        "Eventhouse Name": "string",
        "Eventhouse Id": "string",
        "Description": "string",
    }
    df = _create_dataframe(columns=columns)

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    responses = _base_api(
        request=f"/v1/workspaces/{workspace_id}/eventhouses",
        uses_pagination=True,
        client="fabric_sp",
    )

    for r in responses:
        for v in r.get("value", []):
            new_data = {
                "Eventhouse Name": v.get("displayName"),
                "Eventhouse Id": v.get("id"),
                "Description": v.get("description"),
            }
            df = pd.concat([df, pd.DataFrame(new_data, index=[0])], ignore_index=True)

    return df


def delete_eventhouse(name: str, workspace: Optional[str | UUID] = None):
    """
    Deletes a Fabric eventhouse.

    This is a wrapper function for the following API: `Items - Delete Eventhouse <https://learn.microsoft.com/rest/api/fabric/environment/items/delete-eventhouse>`_.

    Parameters
    ----------
    name: str
        Name of the eventhouse.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    delete_item(item=name, type="Eventhouse", workspace=workspace)


def get_eventhouse_definition(
    eventhouse: str | UUID,
    workspace: Optional[str | UUID] = None,
    return_dataframe: bool = False,
) -> dict | pd.DataFrame:
    """
    Gets the eventhouse definition.

    This is a wrapper function for the following API: `Items - Get Eventhouse Definition <https://learn.microsoft.com/rest/api/fabric/eventhouse/items/get-eventhouse-definition>`_.

    Parameters
    ----------
    eventhouse : str
        Name of the eventhouse.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID in which the eventhouse resides.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    return_dataframe : bool, default=False
        If True, returns a dataframe. If False, returns a json dictionary.

    Returns
    -------
    dict | pandas.DataFrame
        The eventhouse definition in .json format or as a pandas dataframe.
    """

    return get_item_definition(
        item=eventhouse,
        type="Eventhouse",
        workspace=workspace,
        return_dataframe=return_dataframe,
    )



from sempy_labs._helper_functions import (
    resolve_workspace_name_and_id,
    _base_api,
    _create_dataframe,
    _update_dataframe_datatypes,
    create_item,
    delete_item,
)
import pandas as pd
from typing import Optional
from uuid import UUID


def create_sql_database(
    name: str, description: Optional[str] = None, workspace: Optional[str | UUID] = None
):
    """
    Creates a SQL database.

    This is a wrapper function for the following API: `Items - Create SQL Database <https://learn.microsoft.com/rest/api/fabric/sqldatabase/items/create-sql-database>`_.

    Parameters
    ----------
    name: str
        Name of the SQL database.
    description : str, default=None
        A description of the SQL database.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    create_item(
        name=name, description=description, type="SQLDatabase", workspace=workspace
    )


def delete_sql_database(
    sql_database: str | UUID, workspace: Optional[str | UUID] = None
):
    """
    Deletes a SQL Database.

    This is a wrapper function for the following API: `Items - Delete SQL Database <https://learn.microsoft.com/rest/api/fabric/sqldatabase/items/delete-sql-database>`_.

    Parameters
    ----------
    sql_database: str | uuid.UUID
        Name of the SQL database.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    delete_item(item=sql_database, type="SQLDatabase", workspace=workspace)


def list_sql_databases(workspace: Optional[str | UUID] = None) -> pd.DataFrame:
    """
    Lists all SQL databases in the Fabric workspace.

    This is a wrapper function for the following API: `Items - List SQL Databases <https://learn.microsoft.com/rest/api/fabric/sqldatabase/items/list-sql-databases>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of SQL databases in the Fabric workspace.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    columns = {
        "SQL Database Name": "string",
        "SQL Database Id": "string",
        "Description": "string",
        "Connection Info": "string",
        "Database Name": "string",
        "Server FQDN": "string",
    }
    df = _create_dataframe(columns=columns)

    responses = _base_api(
        request=f"/v1/workspaces/{workspace_id}/SQLDatabases",
        uses_pagination=True,
        client="fabric_sp",
    )

    for r in responses:
        for v in r.get("value", []):
            prop = v.get("properties", {})
            new_data = {
                "SQL Database Name": v.get("displayName"),
                "SQL Database Id": v.get("id"),
                "Description": v.get("description"),
                "Connection Info": prop.get("connectionInfo"),
                "Database Name": prop.get("databaseName"),
                "Server FQDN": prop.get("serverFqdn"),
            }

            df = pd.concat([df, pd.DataFrame(new_data, index=[0])], ignore_index=True)

    _update_dataframe_datatypes(dataframe=df, column_map=columns)

    return df


def get_sql_database_tables(
    sql_database: str | UUID, workspace: Optional[str | UUID] = None
) -> pd.DataFrame:
    """
    Shows a list of the tables in the Fabric SQLDabatse. This function is based on INFORMATION_SCHEMA.TABLES.

    Parameters
    ----------
    sql_database : str | uuid.UUID
        Name or ID of the Fabric SQLDabatase.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of the tables in the Fabric SQLDabatase.
    """

    from sempy_labs._sql import ConnectSQLDatabase

    with ConnectSQLDatabase(sql_database=sql_database, workspace=workspace) as sql:
        df = sql.query(
            """
        SELECT TABLE_SCHEMA AS [Schema], TABLE_NAME AS [Table Name], TABLE_TYPE AS [Table Type]
        FROM INFORMATION_SCHEMA.TABLES
        WHERE TABLE_TYPE = 'BASE TABLE'
        """
        )

    return df


def get_sql_database_columns(
    sql_database: str | UUID, workspace: Optional[str | UUID] = None
) -> pd.DataFrame:
    """
    Shows a list of the columns in each table within the Fabric SQLDabatase. This function is based on INFORMATION_SCHEMA.COLUMNS.

    Parameters
    ----------
    sql_database : str | uuid.UUID
        Name or ID of the Fabric SQLDabatase.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of the columns in each table within the Fabric SQLDabatase.
    """

    from sempy_labs._sql import ConnectSQLDatabase

    with ConnectSQLDatabase(sql_database=sql_database, workspace=workspace) as sql:
        df = sql.query(
            """
        SELECT t.TABLE_SCHEMA AS [Schema], t.TABLE_NAME AS [Table Name], c.COLUMN_NAME AS [Column Name], c.DATA_TYPE AS [Data Type], c.IS_NULLABLE AS [Is Nullable], c.CHARACTER_MAXIMUM_LENGTH AS [Character Max Length]
        FROM INFORMATION_SCHEMA.TABLES AS t
        LEFT JOIN INFORMATION_SCHEMA.COLUMNS AS c
        ON t.TABLE_NAME = c.TABLE_NAME
        AND t.TABLE_SCHEMA = c.TABLE_SCHEMA
        WHERE t.TABLE_TYPE = 'BASE TABLE'
        """
        )

    return df



import pandas as pd
from typing import Optional
from sempy_labs._helper_functions import (
    resolve_workspace_name_and_id,
    _base_api,
    delete_item,
    _create_dataframe,
    create_item,
)
from uuid import UUID
import sempy_labs._icons as icons


def list_eventstreams(workspace: Optional[str | UUID] = None) -> pd.DataFrame:
    """
    Shows the eventstreams within a workspace.

    This is a wrapper function for the following API: `Items - List Eventstreams <https://learn.microsoft.com/rest/api/fabric/environment/items/list-eventstreams>`_.

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the eventstreams within a workspace.
    """

    columns = {
        "Eventstream Name": "string",
        "Eventstream Id": "string",
        "Description": "string",
    }
    df = _create_dataframe(columns=columns)

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    responses = _base_api(
        request=f"/v1/workspaces/{workspace_id}/eventstreams", uses_pagination=True
    )

    for r in responses:
        for v in r.get("value", []):
            new_data = {
                "Eventstream Name": v.get("displayName"),
                "Eventstream Id": v.get("id"),
                "Description": v.get("description"),
            }
            df = pd.concat([df, pd.DataFrame(new_data, index=[0])], ignore_index=True)

    return df


def create_eventstream(
    name: str, description: Optional[str] = None, workspace: Optional[str | UUID] = None
):
    """
    Creates a Fabric eventstream.

    This is a wrapper function for the following API: `Items - Create Eventstream <https://learn.microsoft.com/rest/api/fabric/environment/items/create-eventstream>`_.

    Parameters
    ----------
    name: str
        Name of the eventstream.
    description : str, default=None
        A description of the environment.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    create_item(
        name=name, description=description, type="Eventstream", workspace=workspace
    )


def delete_eventstream(
    eventstream: str | UUID, workspace: Optional[str | UUID] = None, **kwargs
):
    """
    Deletes a Fabric eventstream.

    This is a wrapper function for the following API: `Items - Delete Eventstream <https://learn.microsoft.com/rest/api/fabric/environment/items/delete-eventstream>`_.

    Parameters
    ----------
    eventstream: str | uuid.UUID
        Name or ID of the eventstream.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    if "name" in kwargs:
        eventstream = kwargs["name"]
        print(
            f"{icons.warning} The 'name' parameter is deprecated. Please use 'eventstream' instead."
        )

    delete_item(item=eventstream, type="Eventstream", workspace=workspace)



import sempy
import json
import os
import shutil
from sempy_labs.lakehouse._lakehouse import lakehouse_attached
from sempy._utils._log import log
from typing import Optional
import sempy_labs._icons as icons
from uuid import UUID
from sempy_labs._helper_functions import (
    resolve_dataset_name_and_id,
    resolve_workspace_name_and_id,
)


@log
def create_pqt_file(
    dataset: str | UUID,
    workspace: Optional[str | UUID] = None,
    file_name: str = "PowerQueryTemplate",
):
    """
    Dynamically generates a `Power Query Template <https://learn.microsoft.com/power-query/power-query-template>`_ file based on the semantic model. The .pqt file is
    saved within the Files section of your lakehouse.

    Dataflows Gen2 has a `limit of 50 tables <https://learn.microsoft.com/power-query/power-query-online-limits>`_. If there are more than 50 tables, this will save multiple Power Query Template
    files (with each file having a max of 50 tables).

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    file_name : str, default='PowerQueryTemplate'
        The name of the Power Query Template file to be generated.
    """

    sempy.fabric._client._utils._init_analysis_services()
    import Microsoft.AnalysisServices.Tabular as TOM
    from sempy_labs.tom import connect_semantic_model

    if not lakehouse_attached():
        raise ValueError(
            f"{icons.red_dot} In order to run the 'create_pqt_file' function, a lakehouse must be attached to the notebook. Please attach a lakehouse to this notebook."
        )

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (dataset_name, dataset_id) = resolve_dataset_name_and_id(dataset, workspace_id)

    folderPath = "/lakehouse/default/Files"
    subFolderPath = os.path.join(folderPath, "pqtnewfolder")

    with connect_semantic_model(
        dataset=dataset_id, workspace=workspace_id, readonly=True
    ) as tom:
        if not any(
            p.SourceType == TOM.PartitionSourceType.M for p in tom.all_partitions()
        ) and not any(t.RefreshPolicy for t in tom.model.Tables):
            print(
                f"{icons.info} The '{dataset_name}' semantic model within the '{workspace_name}' workspace has no Power Query logic."
            )
            return

        table_map = {}
        expr_map = {}

        for t in tom.model.Tables:
            table_name = t.Name
            for char in icons.special_characters:
                table_name = table_name.replace(char, "")
            if t.RefreshPolicy:
                table_map[table_name] = t.RefreshPolicy.SourceExpression
            elif any(p.SourceType == TOM.PartitionSourceType.M for p in t.Partitions):
                part_name = next(
                    p.Name
                    for p in t.Partitions
                    if p.SourceType == TOM.PartitionSourceType.M
                )
                expr = t.Partitions[part_name].Source.Expression
                table_map[table_name] = expr

        for e in tom.model.Expressions:
            expr_map[e.Name] = [str(e.Kind), e.Expression]

        expressions = tom.model.Expressions.Count

        # Dataflows Gen2 max table limit is 50.
        max_length = 50 - expressions
        table_chunks = [
            dict(list(table_map.items())[i : i + max_length])
            for i in range(0, len(table_map), max_length)
        ]

        def create_pqt(table_map: dict, expr_map: dict, file_name: str):

            os.makedirs(subFolderPath, exist_ok=True)

            class QueryMetadata:
                def __init__(
                    self,
                    QueryName,
                    QueryGroupId=None,
                    LastKnownIsParameter=None,
                    LastKnownResultTypeName=None,
                    LoadEnabled=True,
                    IsHidden=False,
                ):
                    self.QueryName = QueryName
                    self.QueryGroupId = QueryGroupId
                    self.LastKnownIsParameter = LastKnownIsParameter
                    self.LastKnownResultTypeName = LastKnownResultTypeName
                    self.LoadEnabled = LoadEnabled
                    self.IsHidden = IsHidden

            class RootObject:
                def __init__(
                    self,
                    DocumentLocale,
                    EngineVersion,
                    QueriesMetadata,
                    QueryGroups=None,
                ):
                    if QueryGroups is None:
                        QueryGroups = []
                    self.DocumentLocale = DocumentLocale
                    self.EngineVersion = EngineVersion
                    self.QueriesMetadata = QueriesMetadata
                    self.QueryGroups = QueryGroups

            # STEP 1: Create MashupDocument.pq
            mdfileName = "MashupDocument.pq"
            mdFilePath = os.path.join(subFolderPath, mdfileName)
            sb = "section Section1;"
            for t_name, query in table_map.items():
                sb = f'{sb}\nshared #"{t_name}" = '
                if query is not None:
                    pQueryNoSpaces = (
                        query.replace(" ", "")
                        .replace("\n", "")
                        .replace("\t", "")
                        .replace("\r", "")
                    )
                    if pQueryNoSpaces.startswith('letSource=""'):
                        query = 'let\n\tSource = ""\nin\n\tSource'
                sb = f"{sb}{query};"

            for e_name, kind_expr in expr_map.items():
                expr = kind_expr[1]
                sb = f'{sb}\nshared #"{e_name}" = {expr};'

            with open(mdFilePath, "w") as file:
                file.write(sb)

            # STEP 2: Create the MashupMetadata.json file
            mmfileName = "MashupMetadata.json"
            mmFilePath = os.path.join(subFolderPath, mmfileName)
            queryMetadata = []

            for t_name, query in table_map.items():
                queryMetadata.append(
                    QueryMetadata(t_name, None, None, None, True, False)
                )
            for e_name, kind_expr in expr_map.items():
                e_kind = kind_expr[0]
                if e_kind == "M":
                    queryMetadata.append(
                        QueryMetadata(e_name, None, None, None, True, False)
                    )
                else:
                    queryMetadata.append(
                        QueryMetadata(e_name, None, None, None, False, False)
                    )

            rootObject = RootObject(
                "en-US", "2.132.328.0", queryMetadata
            )  # "2.126.453.0"

            def obj_to_dict(obj):
                if isinstance(obj, list):
                    return [obj_to_dict(e) for e in obj]
                elif hasattr(obj, "__dict__"):
                    return {k: obj_to_dict(v) for k, v in obj.__dict__.items()}
                else:
                    return obj

            jsonContent = json.dumps(obj_to_dict(rootObject), indent=4)

            with open(mmFilePath, "w") as json_file:
                json_file.write(jsonContent)

            # STEP 3: Create Metadata.json file
            mFileName = "Metadata.json"
            mFilePath = os.path.join(subFolderPath, mFileName)
            metaData = {"Name": f"{file_name}", "Description": "", "Version": "1.0.0.0"}
            jsonContent = json.dumps(metaData, indent=4)

            with open(mFilePath, "w") as json_file:
                json_file.write(jsonContent)

            # STEP 4: Create [Content_Types].xml file:
            xml_content = """<?xml version="1.0" encoding="utf-8"?><Types xmlns="http://schemas.openxmlformats.org/package/2006/content-types"><Default Extension="json" ContentType="application/json" /><Default Extension="pq" ContentType="application/x-ms-m" /></Types>"""
            xmlFileName = "[Content_Types].xml"
            xmlFilePath = os.path.join(subFolderPath, xmlFileName)
            with open(xmlFilePath, "w", encoding="utf-8") as file:
                file.write(xml_content)

            # STEP 5: Zip up the 4 files
            zipFileName = f"{file_name}.zip"
            zipFilePath = os.path.join(folderPath, zipFileName)
            shutil.make_archive(zipFilePath[:-4], "zip", subFolderPath)

            # STEP 6: Convert the zip file back into a .pqt file
            newExt = ".pqt"
            directory = os.path.dirname(zipFilePath)
            fileNameWithoutExtension = os.path.splitext(os.path.basename(zipFilePath))[
                0
            ]
            newFilePath = os.path.join(directory, fileNameWithoutExtension + newExt)
            shutil.move(zipFilePath, newFilePath)

            # STEP 7: Delete subFolder directory which is no longer needed
            shutil.rmtree(subFolderPath, ignore_errors=True)

            print(
                f"{icons.green_dot} '{file_name}.pqt' has been created based on the '{dataset_name}' semantic model in the '{workspace_name}' workspace within the Files section of your lakehouse."
            )

        a = 0
        for t_map in table_chunks:
            if a > 0:
                save_file_name = f"{file_name}_{a}"
            else:
                save_file_name = file_name
            a += 1
            create_pqt(t_map, expr_map, file_name=save_file_name)



from sempy_labs._helper_functions import (
    resolve_workspace_name_and_id,
    _base_api,
)
from typing import Optional
import sempy_labs._icons as icons
from uuid import UUID


def provision_workspace_identity(workspace: Optional[str | UUID] = None):
    """
    Provisions a workspace identity for a workspace.

    This is a wrapper function for the following API: `Workspaces - Provision Identity <https://learn.microsoft.com/rest/api/fabric/core/workspaces/provision-identity>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    _base_api(
        request=f"/v1/workspaces/{workspace_id}/provisionIdentity",
        method="post",
        lro_return_status_code=True,
        status_codes=None,
        client="fabric_sp",
    )

    print(
        f"{icons.green_dot} A workspace identity has been provisioned for the '{workspace_name}' workspace."
    )


def deprovision_workspace_identity(workspace: Optional[str | UUID] = None):
    """
    Deprovisions a workspace identity for a workspace.

    This is a wrapper function for the following API: `Workspaces - Derovision Identity <https://learn.microsoft.com/rest/api/fabric/core/workspaces/deprovision-identity>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    _base_api(
        request=f"/v1/workspaces/{workspace_id}/deprovisionIdentity",
        method="post",
        lro_return_status_code=True,
        status_codes=None,
        client="fabric_sp",
    )

    print(
        f"{icons.green_dot} The workspace identity has been deprovisioned from the '{workspace_name}' workspace."
    )



import sempy.fabric as fabric
import re
import datetime
from sempy_labs.lakehouse._get_lakehouse_tables import get_lakehouse_tables
from sempy_labs._helper_functions import (
    resolve_lakehouse_name,
    format_dax_object_name,
    retry,
)
from sempy_labs.tom import connect_semantic_model
from typing import Optional
from sempy._utils._log import log
import sempy_labs._icons as icons


@log
def migrate_calc_tables_to_semantic_model(
    dataset: str,
    new_dataset: str,
    workspace: Optional[str] = None,
    new_dataset_workspace: Optional[str] = None,
    lakehouse: Optional[str] = None,
    lakehouse_workspace: Optional[str] = None,
):
    """
    Creates new tables in the Direct Lake semantic model based on the lakehouse tables created using the 'migrate_calc_tables_to_lakehouse' function.

    Parameters
    ----------
    dataset : str
        Name of the import/DirectQuery semantic model.
    new_dataset : str
        Name of the Direct Lake semantic model.
    workspace : str, default=None
        The Fabric workspace name in which the import/DirectQuery semantic model exists.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    new_dataset_workspace : str
        The Fabric workspace name in which the Direct Lake semantic model will be created.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    lakehouse : str, default=None
        The Fabric lakehouse used by the Direct Lake semantic model.
        Defaults to None which resolves to the lakehouse attached to the notebook.
    lakehouse_workspace : str, default=None
        The Fabric workspace used by the lakehouse.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    if dataset == new_dataset:
        raise ValueError(
            f"{icons.red_dot} The 'dataset' and 'new_dataset' parameters are both set to '{dataset}'. These parameters must be set to different values."
        )

    workspace = fabric.resolve_workspace_name(workspace)
    fabric.refresh_tom_cache(workspace=workspace)

    if new_dataset_workspace is None:
        new_dataset_workspace = workspace

    if lakehouse_workspace is None:
        lakehouse_workspace = new_dataset_workspace
    if lakehouse is None:
        lakehouse_id = fabric.get_lakehouse_id()
        lakehouse = resolve_lakehouse_name(lakehouse_id, lakehouse_workspace)

    icons.sll_tags.append("DirectLakeMigration")

    # Get calc tables but not field parameters
    dfP = fabric.list_partitions(dataset=dataset, workspace=workspace)
    dfP_filt = dfP[(dfP["Source Type"] == "Calculated")]
    dfP_filt = dfP_filt[~dfP_filt["Query"].str.contains("NAMEOF")]

    dfC = fabric.list_columns(dataset=dataset, workspace=workspace)
    lc = get_lakehouse_tables(lakehouse=lakehouse, workspace=lakehouse_workspace)
    # Get all calc table columns of calc tables not including field parameters
    dfC_filt = dfC[
        (dfC["Table Name"].isin(dfP_filt["Table Name"]))
    ]  # & (dfC['Type'] == 'CalculatedTableColumn')]
    # dfA = fabric.list_annotations(new_dataset, new_dataset_workspace)
    # dfA_filt = dfA[(dfA['Object Type'] == 'Model') & ~ (dfA['Annotation Value'].str.contains('NAMEOF'))]

    if len(dfP_filt) == 0:
        print(
            f"{icons.green_dot} The '{dataset}' semantic model has no calculated tables."
        )
        return

    @retry(
        sleep_time=1,
        timeout_error_message=f"{icons.red_dot} Function timed out after 1 minute",
    )
    def dyn_connect():
        with connect_semantic_model(
            dataset=new_dataset, readonly=True, workspace=new_dataset_workspace
        ) as tom:

            tom.model

    dyn_connect()

    with connect_semantic_model(
        dataset=new_dataset, readonly=False, workspace=new_dataset_workspace
    ) as tom:
        for tName in dfC_filt["Table Name"].unique():
            if tName.lower() in lc["Table Name"].values:
                if not any(t.Name == tName for t in tom.model.Tables):
                    tom.add_table(name=tName)
                    tom.add_entity_partition(
                        table_name=tName,
                        entity_name=tName.replace(" ", "_").lower(),
                    )

            columns_in_table = dfC_filt.loc[
                dfC_filt["Table Name"] == tName, "Column Name"
            ].unique()

            for cName in columns_in_table:
                scName = dfC.loc[
                    (dfC["Table Name"] == tName) & (dfC["Column Name"] == cName),
                    "Source",
                ].iloc[0]
                cDataType = dfC.loc[
                    (dfC["Table Name"] == tName) & (dfC["Column Name"] == cName),
                    "Data Type",
                ].iloc[0]
                # cType = dfC.loc[
                #    (dfC["Table Name"] == tName)
                #    & (dfC["Column Name"] == cName),
                #    "Type",
                # ].iloc[0]

                # av = tom.get_annotation_value(object = tom.model, name = tName)

                # if cType == 'CalculatedTableColumn':
                # lakeColumn = scName.replace(' ','_')
                # elif cType == 'Calculated':
                pattern = r"\[([^]]+)\]"

                matches = re.findall(pattern, scName)
                lakeColumn = matches[0].replace(" ", "")
                if not any(
                    c.Name == cName and c.Parent.Name == tName
                    for c in tom.all_columns()
                ):
                    tom.add_data_column(
                        table_name=tName,
                        column_name=cName,
                        source_column=lakeColumn,
                        data_type=cDataType,
                    )
                    print(
                        f"{icons.green_dot} The {format_dax_object_name(tName,cName)} column has been added."
                    )

        print(
            f"\n{icons.green_dot} All viable calculated tables have been added to the model."
        )



import sempy.fabric as fabric
import sempy_labs._icons as icons
from typing import Optional
from sempy_labs._helper_functions import (
    resolve_workspace_name_and_id,
    resolve_lakehouse_name_and_id,
    _base_api,
    resolve_item_id,
    _mount,
    resolve_workspace_name,
)
from sempy_labs.lakehouse._lakehouse import lakehouse_attached
from uuid import UUID


def download_report(
    report: str | UUID,
    file_name: Optional[str] = None,
    download_type: str = "LiveConnect",
    workspace: Optional[str | UUID] = None,
):
    """
    Downloads the specified report from the specified workspace to a Power BI .pbix file.

    This is a wrapper function for the following API: `Reports - Export Report In Group <https://learn.microsoft.com/rest/api/power-bi/reports/export-report-in-group>`.

    Parameters
    ----------
    report: str | uuid.UUID
        Name or ID of the report.
    file_name : str, default=None
        Name of the .pbix file to be saved.
        Defaults to None which resolves to the name of the report.
    download_type : str, default="LiveConnect"
        The type of download. Valid values are "LiveConnect" and "IncludeModel".
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    if not lakehouse_attached():
        raise ValueError(
            f"{icons.red_dot} A lakehouse must be attached to the notebook."
        )

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (lakehouse_name, lakehouse_id) = resolve_lakehouse_name_and_id()
    lakehouse_workspace = resolve_workspace_name()

    download_types = ["LiveConnect", "IncludeModel"]
    if download_type not in download_types:
        raise ValueError(
            f"{icons.red_dot} Invalid download_type parameter. Valid options: {download_types}."
        )

    file_name = file_name or report
    report_id = resolve_item_id(item=report, type="Report", workspace=workspace)

    response = _base_api(
        request=f"v1.0/myorg/groups/{workspace_id}/reports/{report_id}/Export?downloadType={download_type}"
    )

    # Save file to the attached lakehouse
    local_path = _mount()
    save_file = f"{local_path}/Files/{file_name}.pbix"
    with open(save_file, "wb") as file:
        file.write(response.content)

    print(
        f"{icons.green_dot} The '{report}' report within the '{workspace_name}' workspace has been exported as the '{file_name}' file in the '{lakehouse_name}' lakehouse within the '{lakehouse_workspace}' workspace."
    )



from sempy_labs._helper_functions import (
    resolve_dataset_id,
    resolve_workspace_name_and_id,
    resolve_report_id,
    _base_api,
    resolve_dataset_name_and_id,
)
from typing import Optional, List
from sempy._utils._log import log
import sempy_labs._icons as icons
from uuid import UUID


@log
def report_rebind(
    report: str | List[str],
    dataset: str,
    report_workspace: Optional[str | UUID] = None,
    dataset_workspace: Optional[str | UUID] = None,
):
    """
    Rebinds a report to a semantic model.

    This is a wrapper function for the following API: `Reports - Rebind Report In Group <https://learn.microsoft.com/rest/api/power-bi/reports/rebind-report-in-group>`_.

    Parameters
    ----------
    report : str | List[str]
        Name(s) of the Power BI report(s).
    dataset : str
        Name of the semantic model.
    report_workspace : str | uuid.UUID, default=None
        The name or ID of the Fabric workspace in which the report resides.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    dataset_workspace : str | uuid.UUID, default=None
        The name or ID of the Fabric workspace in which the semantic model resides.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (report_workspace_name, report_workspace_id) = resolve_workspace_name_and_id(
        report_workspace
    )

    if dataset_workspace is None:
        dataset_workspace = report_workspace_name

    if isinstance(report, str):
        report = [report]

    for rpt in report:
        report_id = resolve_report_id(report=rpt, workspace=report_workspace_id)
        dataset_id = resolve_dataset_id(dataset=dataset, workspace=dataset_workspace)

        payload = {"datasetId": dataset_id}

        _base_api(
            request=f"v1.0/myorg/groups/{report_workspace_id}/reports/{report_id}/Rebind",
            method="post",
            payload=payload,
        )

        print(
            f"{icons.green_dot} The '{rpt}' report has been successfully rebinded to the '{dataset}' semantic model."
        )


@log
def report_rebind_all(
    dataset: str,
    new_dataset: str,
    dataset_workspace: Optional[str] = None,
    new_dataset_workpace: Optional[str] = None,
    report_workspace: Optional[str | List[str]] = None,
):
    """
    Rebinds all reports across all workspaces which are bound to a specific semantic model to a new semantic model.

    Parameters
    ----------
    dataset : str
        Name of the semantic model currently binded to the reports.
    new_dataset : str
        Name of the semantic model to rebind to the reports.
    dataset_workspace : str, default=None
        The name of the Fabric workspace in which the original semantic model resides.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    new_dataset_workspace : str, default=None
        The name of the Fabric workspace in which the new semantic model resides.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    report_workspace : str | List[str], default=None
        The name(s) of the Fabric workspace(s) in which the report(s) reside(s).
        Defaults to None which finds all reports in all workspaces which use the semantic model and rebinds them to
        the new semantic model.
    """

    from sempy_labs._list_functions import list_reports_using_semantic_model

    if dataset == new_dataset:
        raise ValueError(
            f"{icons.red_dot} The 'dataset' and 'new_dataset' parameters are both set to '{dataset}'. These parameters must be set to different values."
        )

    (dataset_name, dataset_id) = resolve_dataset_name_and_id(
        dataset=dataset, workspace=dataset_workspace
    )
    (dataset_workspace_name, dataset_workspace_id) = resolve_workspace_name_and_id(
        workspace=dataset_workspace
    )

    if isinstance(report_workspace, str):
        report_workspace = [report_workspace]

    dfR = list_reports_using_semantic_model(
        dataset=dataset, workspace=dataset_workspace
    )

    if len(dfR) == 0:
        print(
            f"{icons.info} The '{dataset_name}' semantic model within the '{dataset_workspace_name}' workspace has no dependent reports."
        )
        return

    if report_workspace is None:
        dfR_filt = dfR.copy()
    else:
        dfR_filt = dfR[dfR["Report Workspace Name"].isin(report_workspace)]

    for i, r in dfR_filt.iterrows():
        rpt_name = r["Report Name"]
        rpt_wksp = r["Report Workspace Name"]

        report_rebind(
            report=rpt_name,
            dataset=new_dataset,
            report_workspace=rpt_wksp,
            dataset_workspace=new_dataset_workpace,
        )



import pandas as pd
from typing import Optional
import pyarrow.parquet as pq
from sempy_labs._helper_functions import (
    create_abfss_path,
    resolve_workspace_id,
    resolve_lakehouse_id,
    _mount,
)
from sempy._utils._log import log
from tqdm.auto import tqdm
from uuid import UUID
from datetime import datetime


@log
def delta_analyzer_history(
    table_name: str,
    schema: Optional[str] = None,
    lakehouse: Optional[str | UUID] = None,
    workspace: Optional[str | UUID] = None,
) -> pd.DataFrame:
    """
    Analyzes the transaction log for a specified delta table and shows the results in dataframe.  One row per data modification operation.

    Keeps track on the number of Parquet files, rowgroups, file size and #rows impacted by each change.

    Incremental Framing effect: 100% = highly effective, 0% = no benefit at all

    Parameters
    ----------
    table_name : str
        The delta table name.
    schema : str, default=None
        The schema name of the delta table.
    lakehouse : str | uuid.UUID, default=None
        The Fabric lakehouse name or ID.
        Defaults to None which resolves to the lakehouse attached to the notebook.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID used by the lakehouse.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        Displays a gantt visual showing a timeline for individual parquet files.
    """

    import notebookutils
    from IPython.display import display, HTML

    workspace_id = resolve_workspace_id(workspace=workspace)
    lakehouse_id = resolve_lakehouse_id(lakehouse=lakehouse, workspace=workspace)

    table_path = create_abfss_path(lakehouse_id, workspace_id, table_name, schema)
    local_path = _mount(lakehouse=lakehouse, workspace=workspace)
    table_path_local = f"{local_path}/Tables/{table_name}"
    delta_table_path = f"{table_path}/_delta_log"

    files = notebookutils.fs.ls(delta_table_path)
    json_files = [file.name for file in files if file.name.endswith(".json")]

    element_version = total_size = total_rows = total_files = total_rowgroups = 0
    changes_array = []
    parquet_files = []
    my_date_time_format = "%Y-%m-%d %H:%M:%S.%f"
    now_to_epoch = datetime.now().strftime(my_date_time_format)
    num_latest_files = len(json_files)

    for idx, file in enumerate(bar := tqdm(json_files), start=1):
        bar.set_description(
            f"Analyzing the '{file}' parquet file ({idx}/{num_latest_files})..."
        )

        change_timestamp = datetime.strptime(
            "2001-01-01 12:00:00.000", my_date_time_format
        )
        df = pd.read_json(f"{delta_table_path}/{file}", lines=True)

        rows_added = size_added = rows_deleted = size_deleted = files_added = (
            files_removed
        ) = row_groups_added = row_groups_removed = 0
        total_files_before_change = total_files
        total_row_groups_before_change = total_rowgroups
        operation = predicate = tags = ""

        for _, row in df.iterrows():
            add_row = row.get("add")
            remove_row = row.get("remove")
            commit_row = row.get("commitInfo")

            if isinstance(add_row, dict):
                file_name = add_row["path"]
                fs_filename = f"{table_path}/{file_name}"
                size_added += add_row["size"]
                files_added += 1
                filerows_added = 0

                if notebookutils.fs.exists(fs_filename):
                    parquet_file = pq.ParquetFile(table_path_local + f"/{file_name}")
                    for i in range(parquet_file.num_row_groups):
                        row_group = parquet_file.metadata.row_group(i)
                        num_rows = row_group.num_rows
                        filerows_added += num_rows
                        rows_added += num_rows

                    row_groups_added += parquet_file.num_row_groups

                    start = str(
                        datetime.fromtimestamp(add_row["modificationTime"] / 1000.0)
                    )
                    parquet_files.append(
                        {
                            "file": file_name,
                            "start": start,
                            "end": now_to_epoch,
                            "rows": filerows_added,
                            "isCurrent": 1,
                        }
                    )

            if isinstance(remove_row, dict):
                file_name = remove_row["path"]
                fs_filename = f"{table_path}/{file_name}"

                if notebookutils.fs.exists(fs_filename):
                    parquet_file = pq.ParquetFile(table_path_local + f"/{file_name}")
                    for i in range(parquet_file.num_row_groups):
                        row_group = parquet_file.metadata.row_group(i)
                        num_rows = row_group.num_rows
                        rows_deleted += num_rows

                    files_removed += 1
                    size_deleted += remove_row.get("size", 0)
                    row_groups_removed += parquet_file.num_row_groups

                    result = next(
                        (row for row in parquet_files if row["file"] == file_name), None
                    )
                    if result:
                        result.update(
                            {
                                "isCurrent": 0,
                                "end": str(
                                    datetime.fromtimestamp(
                                        remove_row["deletionTimestamp"] / 1000.0
                                    )
                                ),
                            }
                        )

            if isinstance(commit_row, dict):
                operation = commit_row.get("operation")
                tags = commit_row.get("tags")
                predicate = commit_row.get("operationParameters", {}).get("predicate")

                if operation == "VACUUM START":
                    operation_metrics = commit_row.get("operationMetrics", {})
                    total_files -= int(operation_metrics.get("numFilesToDelete", 0))
                    total_size -= int(operation_metrics.get("sizeOfDataToDelete", 0))

                change_timestamp = datetime.fromtimestamp(
                    commit_row["timestamp"] / 1000.0
                )

        total_size += size_added - size_deleted
        total_rows += rows_added - rows_deleted
        total_files += files_added - files_removed
        total_rowgroups += row_groups_added - row_groups_removed

        incremental_framing_effect = 1
        if size_deleted != 0:
            incremental_framing_effect = (
                int((total_size - size_added * 1.0) / total_size * 100000) / 1000
            )
            # incrementalFramingEffect = round(
            #     (totalSize - sizeAdded * 1.0) / totalSize, 4
            # )

        changes_array.append(
            [
                element_version,
                operation,
                predicate,
                change_timestamp,
                incremental_framing_effect,
                files_added,
                files_removed,
                total_files_before_change - files_removed,
                total_files,
                size_added,
                size_deleted,
                total_size,
                row_groups_added,
                row_groups_removed,
                total_row_groups_before_change - row_groups_removed,
                total_rowgroups,
                rows_added,
                rows_deleted,
                rows_added - rows_deleted,
                total_rows,
                tags,
            ]
        )

        element_version += 1

    #  /********************************************************************************************************************
    #      Display Gantt Chart of files
    #  ********************************************************************************************************************/
    spec: str = (
        """{
    "$$schema": 'https://vega.github.io/schema/vega-lite/v2.json',
    "description": "A simple bar chart with ranged data (aka Gantt Chart).",
    "width" : 1024 ,
    "data": {
        "values": %s
    },
    "layer":[
        {"mark": "bar"},
        {"mark": {
        "type": "text",
        "align": "center",
        "baseline": "middle",
        "dx": 40
        },
        "encoding": {
        "text": {"field": "rows", "type": "quantitative", "format":","},
        "color":{
        "condition": {"test": "datum['isCurrent'] == 1", "value": "black"},
        "value": "black"
            }
        }
        }],
    "encoding": {
        "y": {"field": "file", "type": "ordinal","sort": "isCurrent","title":null,"axis":{"labelPadding":15,"labelLimit":360}},
        "x": {"field": "start", "type": "temporal","title":null},
        "x2": {"field": "end", "type": "temporal","title":null},
            "color": {
            "field": "isCurrent",
            "scale": {"range": ["silver", "#ca8861"]}
            }
    }
    }"""
        % (parquet_files)
    )

    display(
        HTML(
            """
        <!DOCTYPE html>
        <html>
            <head>
                <script src="https://cdn.jsdelivr.net/npm/vega@5"></script>
                <script src="https://cdn.jsdelivr.net/npm/vega-lite@5"></script>
                <script src="https://cdn.jsdelivr.net/npm/vega-embed@6"></script>
            </head>
            <body>
                <div id="vis"></div>
                <script type="text/javascript">
                    var spec = """
            + spec
            + """;
                    var opt = {"renderer": "canvas", "actions": false};
                    vegaEmbed("#vis", spec, opt);
                </script>
            </body>
        </html>"""
        )
    )

    return pd.DataFrame(
        changes_array,
        columns=[
            "Change Number",
            "Change Type",
            "Predicate",
            "Modification Time",
            "Incremental Effect",
            "Files Added",
            "Files Removed",
            "Files Preserved",
            "Files After Change",
            "Size Added",
            "Sized Removed",
            "Size After Change",
            "Rowgroups Added",
            "Rowgroups Removed",
            "Rowgroups Preserved",
            "Rowgroups After Change",
            "Rows Added",
            "Rows Removed",
            "Rows Delta",
            "Rows After Change",
            "Tags",
        ],
    )



import sempy.fabric as fabric
import pandas as pd
from typing import Optional
from sempy_labs._list_functions import list_semantic_model_objects
from sempy._utils._log import log
import sempy_labs._icons as icons


@log
def migration_validation(
    dataset: str,
    new_dataset: str,
    workspace: Optional[str] = None,
    new_dataset_workspace: Optional[str] = None,
) -> pd.DataFrame:
    """
    Shows the objects in the original semantic model and whether then were migrated successfully or not.

    Parameters
    ----------
    dataset : str
        Name of the import/DirectQuery semantic model.
    new_dataset : str
        Name of the Direct Lake semantic model.
    workspace : str, default=None
        The Fabric workspace name in which the import/DirectQuery semantic model exists.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    new_dataset_workspace : str
        The Fabric workspace name in which the Direct Lake semantic model will be created.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
       A pandas dataframe showing a list of objects and whether they were successfully migrated. Also shows the % of objects which were migrated successfully.
    """

    if dataset == new_dataset:
        raise ValueError(
            f"{icons.red_dot} The 'dataset' and 'new_dataset' parameters are both set to '{dataset}'. These parameters must be set to different values."
        )

    icons.sll_tags.append("DirectLakeMigration")

    dfA = list_semantic_model_objects(dataset=dataset, workspace=workspace)
    dfB = list_semantic_model_objects(
        dataset=new_dataset, workspace=new_dataset_workspace
    )

    def is_migrated(row):
        if row["Object Type"] == "Calculated Table":
            return (
                (dfB["Parent Name"] == row["Parent Name"])
                & (dfB["Object Name"] == row["Object Name"])
                & (dfB["Object Type"].isin(["Calculated Table", "Table"]))
            ).any()
        else:
            return (
                (dfB["Parent Name"] == row["Parent Name"])
                & (dfB["Object Name"] == row["Object Name"])
                & (dfB["Object Type"] == row["Object Type"])
            ).any()

    dfA["Migrated"] = dfA.apply(is_migrated, axis=1)

    denom = len(dfA)
    num = len(dfA[dfA["Migrated"]])
    print(f"{100 * round(num / denom,2)}% migrated")

    return dfA



import pandas as pd
from uuid import UUID
import sempy_labs._icons as icons
from typing import Optional, Tuple
from sempy._utils._log import log
from sempy_labs._helper_functions import (
    _base_api,
    _create_dataframe,
    _update_dataframe_datatypes,
    _is_valid_uuid,
    get_capacity_id,
)


def patch_capacity(capacity: str | UUID, tenant_key_id: UUID):
    """
    Changes specific capacity information. Currently, this API call only supports changing the capacity's encryption key.

    This is a wrapper function for the following API: `Admin - Patch Capacity As Admin <https://learn.microsoft.com/rest/api/power-bi/admin/patch-capacity-as-admin>`_.

    Parameters
    ----------
    capacity : str | uuid.UUID
        The name or ID of the capacity.
    tenant_key_id : str
        The ID of the encryption key.
    """

    (capacity_name, capacity_id) = _resolve_capacity_name_and_id(capacity)

    payload = {
        "tenantKeyId": tenant_key_id,
    }

    _base_api(
        request=f"/v1.0/myorg/admin/capacities/{capacity_id}",
        method="patch",
        payload=payload,
    )

    print(
        f"{icons.green_dot} The capacity '{capacity_name}' has been successfully patched."
    )


def _resolve_capacity_name_and_id(
    capacity: str | UUID,
) -> Tuple[str, UUID]:

    dfC = list_capacities(capacity=capacity)
    if dfC.empty:
        raise ValueError(f"{icons.red_dot} The '{capacity}' capacity was not found.")

    capacity_name = dfC["Capacity Name"].iloc[0]
    capacity_id = dfC["Capacity Id"].iloc[0]

    return capacity_name, capacity_id


def _list_capacities_meta() -> pd.DataFrame:
    """
    Shows the a list of capacities and their properties. This function is the admin version.

    This is a wrapper function for the following API: `Admin - Get Capacities As Admin <https://learn.microsoft.com/rest/api/power-bi/admin/get-capacities-as-admin>`_.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the capacities and their properties
    """

    columns = {
        "Capacity Id": "string",
        "Capacity Name": "string",
        "Sku": "string",
        "Region": "string",
        "State": "string",
        "Admins": "string",
    }
    df = _create_dataframe(columns=columns)

    responses = _base_api(
        request="/v1.0/myorg/admin/capacities", client="fabric_sp", uses_pagination=True
    )

    for r in responses:
        for i in r.get("value", []):
            new_data = {
                "Capacity Id": i.get("id").lower(),
                "Capacity Name": i.get("displayName"),
                "Sku": i.get("sku"),
                "Region": i.get("region"),
                "State": i.get("state"),
                "Admins": [i.get("admins", [])],
            }
            df = pd.concat([df, pd.DataFrame(new_data, index=[0])], ignore_index=True)

    return df


def get_capacity_assignment_status(
    workspace: Optional[str | UUID] = None,
) -> pd.DataFrame:
    """
    Gets the status of the assignment-to-capacity operation for the specified workspace.

    This is a wrapper function for the following API: `Capacities - Groups CapacityAssignmentStatus <https://learn.microsoft.com/rest/api/power-bi/capacities/groups-capacity-assignment-status>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the status of the assignment-to-capacity operation for the specified workspace.
    """
    from sempy_labs.admin._basic_functions import _resolve_workspace_name_and_id

    (workspace_name, workspace_id) = _resolve_workspace_name_and_id(workspace)

    columns = {
        "Status": "string",
        "Activity Id": "string",
        "Start Time": "datetime",
        "End Time": "datetime",
        "Capacity Id": "string",
        "Capacity Name": "string",
    }
    df = _create_dataframe(columns=columns)

    response = _base_api(
        request=f"/v1.0/myorg/groups/{workspace_id}/CapacityAssignmentStatus",
        client="fabric_sp",
    )
    v = response.json()
    capacity_id = v.get("capacityId")

    (capacity_name, capacity_id) = _resolve_capacity_name_and_id(capacity=capacity_id)

    new_data = {
        "Status": v.get("status"),
        "Activity Id": v.get("activityId"),
        "Start Time": v.get("startTime"),
        "End Time": v.get("endTime"),
        "Capacity Id": capacity_id,
        "Capacity Name": capacity_name,
    }

    df = pd.concat([df, pd.DataFrame([new_data])], ignore_index=True)

    _update_dataframe_datatypes(dataframe=df, column_map=columns)

    return df


def get_capacity_state(capacity: Optional[str | UUID] = None):
    """
    Gets the state of a capacity.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    capacity : str | uuid.UUID, default=None
        The capacity name or ID.
        Defaults to None which resolves to the capacity of the attached lakehouse
        or if no lakehouse is attached, resolves to the workspace of the notebook.

    Returns
    -------
    str
        The capacity state.
    """

    df = list_capacities()

    if capacity is None:
        capacity = get_capacity_id()
    if _is_valid_uuid(capacity):
        df_filt = df[df["Capacity Id"] == capacity]
    else:
        df_filt = df[df["Capacity Name"] == capacity]

    if df_filt.empty:
        raise ValueError(f"{icons.red_dot} The capacity '{capacity}' was not found.")

    return df_filt["State"].iloc[0]


@log
def list_capacities(
    capacity: Optional[str | UUID] = None,
) -> pd.DataFrame:
    """
    Shows the a list of capacities and their properties.

    This is a wrapper function for the following API: `Admin - Get Capacities As Admin <https://learn.microsoft.com/rest/api/power-bi/admin/get-capacities-as-admin>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    capacity : str | uuid.UUID, default=None
        The capacity name or ID to filter. If None, all capacities are returned.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the capacities and their properties.
    """

    columns = {
        "Capacity Id": "string",
        "Capacity Name": "string",
        "Sku": "string",
        "Region": "string",
        "State": "string",
        "Admins": "string",
    }
    df = _create_dataframe(columns=columns)

    responses = _base_api(
        request="/v1.0/myorg/admin/capacities", client="fabric_sp", uses_pagination=True
    )

    for r in responses:
        for i in r.get("value", []):
            new_data = {
                "Capacity Id": i.get("id").lower(),
                "Capacity Name": i.get("displayName"),
                "Sku": i.get("sku"),
                "Region": i.get("region"),
                "State": i.get("state"),
                "Admins": [i.get("admins", [])],
            }
            df = pd.concat([df, pd.DataFrame(new_data, index=[0])], ignore_index=True)

    if capacity is not None:
        if _is_valid_uuid(capacity):
            df = df[df["Capacity Id"] == capacity.lower()]
        else:
            df = df[df["Capacity Name"] == capacity]

    return df


def list_capacity_users(capacity: str | UUID) -> pd.DataFrame:
    """
    Shows a list of users that have access to the specified workspace.

    This is a wrapper function for the following API: `Admin - Capacities GetCapacityUsersAsAdmin <https://learn.microsoft.com/rest/api/power-bi/admin/capacities-get-capacity-users-as-admin>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    capacity : str | uuid.UUID
        The name or ID of the capacity.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of users that have access to the specified workspace.
    """

    (capacity_name, capacity_id) = _resolve_capacity_name_and_id(capacity)

    columns = {
        "User Name": "string",
        "Email Address": "string",
        "Capacity User Access Right": "string",
        "Identifier": "string",
        "Graph Id": "string",
        "Principal Type": "string",
        "User Type": "string",
        "Profile": "string",
    }

    df = _create_dataframe(columns=columns)

    response = _base_api(
        request=f"/v1.0/myorg/admin/capacities/{capacity_id}/users", client="fabric_sp"
    )

    rows = []
    for v in response.json().get("value", []):
        rows.append(
            {
                "User Name": v.get("displayName"),
                "Email Address": v.get("emailAddress"),
                "Capacity User Access Right": v.get("capacityUserAccessRight"),
                "Identifier": v.get("identifier"),
                "Graph Id": v.get("graphId"),
                "Principal Type": v.get("principalType"),
                "User Type": v.get("userType"),
                "Profile": v.get("profile"),
            }
        )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))

    _update_dataframe_datatypes(dataframe=df, column_map=columns)

    return df



import pandas as pd
import json
from typing import Optional
from sempy_labs._helper_functions import (
    resolve_workspace_name_and_id,
    _base_api,
    _create_dataframe,
    _update_dataframe_datatypes,
    resolve_item_id,
    _decode_b64,
    delete_item,
    get_item_definition,
)

from uuid import UUID


def list_mounted_data_factories(
    workspace: Optional[str | UUID] = None,
) -> pd.DataFrame:
    """
    Shows a list of mounted data factories from the specified workspace.

    This is a wrapper function for the following API: `Items - List Mounted Data Factories <https://learn.microsoft.com/rest/api/fabric/mounteddatafactory/items/list-mounted-data-factories>`_.

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of mounted data factories from the specified workspace.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    columns = {
        "Mounted Data Factory Name": "str",
        "Mounted Data Factory Id": "str",
        "Description": "str",
    }

    df = _create_dataframe(columns=columns)
    responses = _base_api(
        request=f"/v1/workspaces/{workspace_id}/mountedDataFactories",
        uses_pagination=True,
    )

    for r in responses:
        for v in r.get("value", []):
            new_data = {
                "Mounted Data Factory Name": v.get("displayName"),
                "Mounted Data Factory Id": v.get("id"),
                "Description": v.get("description"),
            }

            df = pd.concat([df, pd.DataFrame(new_data, index=[0])], ignore_index=True)

    _update_dataframe_datatypes(dataframe=df, column_map=columns)

    return df


def get_mounted_data_factory_definition(
    mounted_data_factory: str | UUID, workspace: Optional[str | UUID] = None
) -> dict:
    """
    Returns the specified MountedDataFactory public definition.

    This is a wrapper function for the following API: `Items - Get Mounted Data Factory Definition <https://learn.microsoft.com/rest/api/fabric/mounteddatafactory/items/get-mounted-data-factory-definition>`_.

    Parameters
    ----------
    mounted_data_factory : str | uuid.UUID
        The name or ID of the mounted data factory.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    dict
        The 'mountedDataFactory-content.json' file from the mounted data factory definition.
    """

    return get_item_definition(
        item=mounted_data_factory,
        type="MountedDataFactory",
        workspace=workspace,
        return_dataframe=False,
    )


def delete_mounted_data_factory(
    mounted_data_factory: str | UUID, workspace: Optional[str | UUID]
):
    """
    Deletes the specified mounted data factory.

    This is a wrapper function for the following API: `Items - Delete Mounted Data Factory <https://learn.microsoft.com/rest/api/fabric/mounteddatafactory/items/delete-mounted-data-factory>`_.

    Parameters
    ----------
    mounted_data_factory : str | uuid.UUID
        The name or ID of the mounted data factory.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    delete_item(
        item=mounted_data_factory, type="MountedDataFactory", workspace=workspace
    )



import sempy
import pandas as pd
from sempy_labs.lakehouse import get_lakehouse_columns
from sempy_labs.directlake._dl_helper import get_direct_lake_source
from sempy_labs.tom import connect_semantic_model
from sempy_labs._helper_functions import (
    _convert_data_type,
    resolve_workspace_name_and_id,
    resolve_dataset_name_and_id,
)
from typing import Optional
from sempy._utils._log import log
import sempy_labs._icons as icons
from uuid import UUID


@log
def direct_lake_schema_sync(
    dataset: str | UUID,
    workspace: Optional[str | UUID] = None,
    add_to_model: bool = False,
    remove_from_model: bool = False,
) -> pd.DataFrame:
    """
    Shows/adds columns which exist in the lakehouse but do not exist in the semantic model (only for tables in the semantic model).

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    add_to_model : bool, default=False
        If set to True, columns which exist in the lakehouse but do not exist in the semantic model are added to the semantic model. No new tables are added.
    remove_from_model : bool, default=False
        If set to True, columns which exist in the semantic model but do not exist in the lakehouse are removed from the semantic model. No new tables are removed.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the status of columns in the semantic model and lakehouse (prior to adding/removing them from the model using this function).
    """

    sempy.fabric._client._utils._init_analysis_services()
    import Microsoft.AnalysisServices.Tabular as TOM

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (dataset_name, dataset_id) = resolve_dataset_name_and_id(dataset, workspace_id)

    artifact_type, lakehouse_name, lakehouse_id, lakehouse_workspace_id = (
        get_direct_lake_source(dataset=dataset_id, workspace=workspace_id)
    )

    if artifact_type == "Warehouse":
        raise ValueError(
            f"{icons.red_dot} This function is only valid for Direct Lake semantic models which source from Fabric lakehouses (not warehouses)."
        )

    if artifact_type is None:
        raise ValueError(
            f"{icons.red_dot} This function only supports Direct Lake semantic models where the source lakehouse resides in the same workpace as the semantic model."
        )

    lc = get_lakehouse_columns(lakehouse_id, lakehouse_workspace_id)

    readonly = True
    if add_to_model or remove_from_model:
        readonly = False
    df = pd.DataFrame(
        columns=[
            "TableName",
            "ColumnName",
            "SourceTableName",
            "SourceColumnName",
            "Status",
        ]
    )

    with connect_semantic_model(
        dataset=dataset_id, readonly=readonly, workspace=workspace_id
    ) as tom:
        # Check if the columns in the semantic model exist in the lakehouse
        for c in tom.all_columns():
            partition_name = next(p.Name for p in c.Table.Partitions)
            p = c.Table.Partitions[partition_name]
            if p.SourceType == TOM.PartitionSourceType.Entity:
                entity_name = p.Source.EntityName
                source_column = c.SourceColumn
                lc_filt = lc[
                    (lc["Table Name"] == entity_name)
                    & (lc["Column Name"] == source_column)
                ]
                # Remove column from model if it doesn't exist in the lakehouse
                if lc_filt.empty:
                    new_data = {
                        "TableName": c.Parent.Name,
                        "ColumnName": c.Name,
                        "SourceTableName": entity_name,
                        "SourceColumnName": source_column,
                        "Status": "Not in lakehouse",
                    }
                    df = pd.concat(
                        [df, pd.DataFrame(new_data, index=[0])], ignore_index=True
                    )
                    if remove_from_model:
                        tom.remove_object(object=c)
                        print(
                            f"{icons.green_dot} The '{c.Parent.Name}'[{c.Name}] column has been removed from the '{dataset_name}' semantic model within the '{workspace_name}' workspace."
                        )

        # Check if the lakehouse columns exist in the semantic model
        for i, r in lc.iterrows():
            lakeTName = r["Table Name"]
            lakeCName = r["Column Name"]
            dType = r["Data Type"]

            if any(
                p.Source.EntityName == lakeTName
                for p in tom.all_partitions()
                if p.SourceType == TOM.PartitionSourceType.Entity
            ):
                table_name = next(
                    t.Name
                    for t in tom.model.Tables
                    for p in t.Partitions
                    if p.SourceType == TOM.PartitionSourceType.Entity
                    and p.Source.EntityName == lakeTName
                )

                if not any(
                    c.SourceColumn == lakeCName and c.Parent.Name == table_name
                    for c in tom.all_columns()
                ):
                    new_data = {
                        "TableName": table_name,
                        "ColumnName": None,
                        "SourceTableName": lakeTName,
                        "SourceColumnName": lakeCName,
                        "Status": "Not in semantic model",
                    }
                    df = pd.concat(
                        [df, pd.DataFrame(new_data, index=[0])], ignore_index=True
                    )

                    if add_to_model:
                        dt = _convert_data_type(dType)
                        tom.add_data_column(
                            table_name=table_name,
                            column_name=lakeCName,
                            source_column=lakeCName,
                            data_type=dt,
                        )
                        print(
                            f"{icons.green_dot} The '{lakeCName}' column in the '{lakeTName}' lakehouse table was added to the '{dataset_name}' semantic model within the '{workspace_name}' workspace."
                        )

        return df



import pandas as pd
from typing import Optional
import sempy_labs._icons as icons
from sempy_labs._helper_functions import (
    _update_dataframe_datatypes,
    _base_api,
    _create_dataframe,
)
from uuid import UUID


def list_workloads(capacity: str | UUID, **kwargs) -> pd.DataFrame:
    """
    Returns the current state of the specified capacity workloads.
    If a workload is enabled, the percentage of maximum memory that the workload can consume is also returned.

    This is a wrapper function for the following API: `Capacities - Get Workloads <https://learn.microsoft.com/rest/api/power-bi/capacities/get-workloads>`_.

    Parameters
    ----------
    capacity : str | uuid.UUID
        The capacity name or ID.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the current state of the specified capacity workloads.
    """

    from sempy_labs._helper_functions import resolve_capacity_id

    if "capacity_name" in kwargs:
        capacity = kwargs["capacity_name"]
        print(
            f"{icons.warning} The 'capacity_name' parameter is deprecated. Please use 'capacity' instead."
        )

    columns = {
        "Workload Name": "string",
        "State": "string",
        "Max Memory Percentage Set By User": "int",
    }
    df = _create_dataframe(columns=columns)

    capacity_id = resolve_capacity_id(capacity=capacity)

    response = _base_api(request=f"/v1.0/myorg/capacities/{capacity_id}/Workloads")

    for v in response.json().get("value", []):
        new_data = {
            "Workload Name": v.get("name"),
            "State": v.get("state"),
            "Max Memory Percentage Set By User": v.get("maxMemoryPercentageSetByUser"),
        }
        df = pd.concat([df, pd.DataFrame(new_data, index=[0])], ignore_index=True)

    _update_dataframe_datatypes(dataframe=df, column_map=columns)

    return df


def patch_workload(
    capacity: str | UUID,
    workload_name: str,
    state: Optional[str] = None,
    max_memory_percentage: Optional[int] = None,
    **kwargs,
):
    """
    Changes the state of a specific workload to Enabled or Disabled.
    When enabling a workload, specify the percentage of maximum memory that the workload can consume.

    This is a wrapper function for the following API: `Capacities - Patch Workload <https://learn.microsoft.com/rest/api/power-bi/capacities/patch-workload>`_.

    Parameters
    ----------
    capacity : str | uuid.UUID
        The capacity name or ID.
    workload_name : str
        The workload name.
    state : str, default=None
        The capacity workload state.
    max_memory_percentage : int, default=None
        The percentage of the maximum memory that a workload can consume (set by the user).
    """

    from sempy_labs._helper_functions import resolve_capacity_id

    if "capacity_name" in kwargs:
        capacity = kwargs["capacity_name"]
        print(
            f"{icons.warning} The 'capacity_name' parameter is deprecated. Please use 'capacity' instead."
        )

    capacity_id = resolve_capacity_id(capacity=capacity)

    states = ["Disabled", "Enabled", "Unsupported"]
    state = state.capitalize()
    if state is not None and state not in states:
        raise ValueError(
            f"{icons.red_dot} Invalid 'state' parameter. Please choose from these options: {states}."
        )
    if max_memory_percentage is not None and (
        max_memory_percentage < 0 or max_memory_percentage > 100
    ):
        raise ValueError(
            f"{icons.red_dot} Invalid max memory percentage. Must be a value between 0-100."
        )

    url = f"/v1.0/myorg/capacities/{capacity_id}/Workloads/{workload_name}"

    get_response = _base_api(request=url)
    get_json = get_response.json().get("value")
    current_state = get_json.get("state")
    current_max_memory = get_json.get("maxMemoryPercentageSetByUser")

    if current_state == state and str(current_max_memory) == str(max_memory_percentage):
        print(
            f"{icons.info} The current workload settings are the same as those specified in the parameters of this function. The workload has not been updated."
        )
        return

    payload = {}
    if state is not None:
        payload["state"] = state
    else:
        payload["state"] = current_state
    if max_memory_percentage is not None:
        payload["maxMemoryPercentageSetByUser"] = max_memory_percentage
    else:
        payload["maxMemoryPercentageSetByUser"] = current_max_memory

    _base_api(request=url, method="patch", payload=payload)

    print(
        f"The '{workload_name}' workload within the '{capacity}' capacity has been updated accordingly."
    )



import pandas as pd
from typing import Optional
from sempy_labs._helper_functions import (
    resolve_workspace_name_and_id,
    _base_api,
    delete_item,
    _create_dataframe,
    create_item,
)
from uuid import UUID


def list_ml_models(workspace: Optional[str | UUID] = None) -> pd.DataFrame:
    """
    Shows the ML models within a workspace.

    This is a wrapper function for the following API: `Items - List ML Models <https://learn.microsoft.com/rest/api/fabric/mlmodel/items/list-ml-models>`_.

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the ML models within a workspace.
    """

    columns = {
        "ML Model Name": "string",
        "ML Model Id": "string",
        "Description": "string",
    }
    df = _create_dataframe(columns=columns)

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    responses = _base_api(
        request=f"/v1/workspaces/{workspace_id}/mlModels",
        status_codes=200,
        uses_pagination=True,
    )

    for r in responses:
        for v in r.get("value", []):
            model_id = v.get("id")
            modelName = v.get("displayName")
            desc = v.get("description")

            new_data = {
                "ML Model Name": modelName,
                "ML Model Id": model_id,
                "Description": desc,
            }
            df = pd.concat([df, pd.DataFrame(new_data, index=[0])], ignore_index=True)

    return df


def create_ml_model(
    name: str, description: Optional[str] = None, workspace: Optional[str | UUID] = None
):
    """
    Creates a Fabric ML model.

    This is a wrapper function for the following API: `Items - Create ML Model <https://learn.microsoft.com/rest/api/fabric/mlmodel/items/create-ml-model>`_.

    Parameters
    ----------
    name: str
        Name of the ML model.
    description : str, default=None
        A description of the ML model.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    create_item(name=name, description=description, type="MLModel", workspace=workspace)


def delete_ml_model(name: str | UUID, workspace: Optional[str | UUID] = None):
    """
    Deletes a Fabric ML model.

    This is a wrapper function for the following API: `Items - Delete ML Model <https://learn.microsoft.com/rest/api/fabric/mlmodel/items/delete-ml-model>`_.

    Parameters
    ----------
    name: str | uuid.UUID
        Name or ID of the ML model.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    delete_item(item=name, type="MLModel", workspace=workspace)



item_types = {
    "Dashboard": ["Dashboard", "dashboards"],
    "DataPipeline": ["Data Pipeline", "dataPipelines", "pipeline-content.json"],
    "Datamart": ["Datamart", "datamarts"],
    "Environment": ["Environment", "environments"],
    "Eventhouse": ["Eventhouse", "eventhouses", "EventhouseProperties.json"],
    "Eventstream": ["Eventstream", "eventstreams", "eventstream.json"],
    "GraphQLApi": ["GraphQL Api", "GraphQLApis"],
    "KQLDashboard": ["KQL Dashboard", "kqlDashboards", "RealTimeDashboard.json"],
    "KQLDatabase": [
        "KQL Database",
        "kqlDatabases",
    ],  # "DatabaseProperties.json", "DatabaseSchema.kql"
    "KQLQueryset": ["KQL Queryset", "kqlQuerysets", "RealTimeQueryset.json"],
    "Lakehouse": ["Lakehouse", "lakehouses"],
    "MLExperiment": ["ML Experiment", "mlExperiments"],
    "MLModel": ["ML Model", "mlModels"],
    "MirroredDatabase": [
        "Mirrored Database",
        "mirroredDatabases",
        "mirroredDatabase.json",
    ],
    "MirroredWarehouse": ["Mirrored Warehouse", "mirroredWarehouses"],
    "MountedDataFactory": [
        "Mounted Data Factory",
        "mountedDataFactories",
        "mountedDataFactory-content.json",
    ],
    "Notebook": ["Notebook", "notebooks"],
    "PaginatedReport": ["Paginated Report", "paginatedReports"],
    "Reflex": ["Reflex", "reflexes", "ReflexEntities.json"],
    "Report": ["Report", "reports", "report.json"],
    "SQLDatabase": ["SQL Database", "sqlDatabases"],
    "SQLEndpoint": ["SQL Endpoint", "sqlEndpoints"],
    "SemanticModel": ["Semantic Model", "semanticModels", "model.bim"],
    "SparkJobDefinition": [
        "Spark Job Definition",
        "sparkJobDefinitions",
        "SparkJobDefinitionV1.json",
    ],
    "Warehouse": ["Warehouse", "warehouses"],
}



import sempy.fabric as fabric
import pandas as pd
from sempy_labs._helper_functions import (
    resolve_workspace_name_and_id,
    resolve_dataset_name_and_id,
    _update_dataframe_datatypes,
    _base_api,
    _create_dataframe,
)
from sempy._utils._log import log
from typing import Optional, Tuple
import sempy_labs._icons as icons
from uuid import UUID


@log
def qso_sync(dataset: str | UUID, workspace: Optional[str | UUID] = None):
    """
    Triggers a query scale-out sync of read-only replicas for the specified dataset from the specified workspace.

    This is a wrapper function for the following API: `Datasets - Trigger Query Scale Out Sync In Group <https://learn.microsoft.com/rest/api/power-bi/datasets/trigger-query-scale-out-sync-in-group>`_.

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (dataset_name, dataset_id) = resolve_dataset_name_and_id(dataset, workspace_id)

    _base_api(
        request=f"/v1.0/myorg/groups/{workspace_id}/datasets/{dataset_id}/queryScaleOut/sync",
        method="post",
    )
    print(
        f"{icons.green_dot} QSO sync initiated for the '{dataset_name}' semantic model within the '{workspace_name}' workspace."
    )


@log
def qso_sync_status(
    dataset: str | UUID, workspace: Optional[str | UUID] = None
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """
    Returns the query scale-out sync status for the specified dataset from the specified workspace.

    This is a wrapper function for the following API: `Datasets - Get Query Scale Out Sync Status In Group <https://learn.microsoft.com/rest/api/power-bi/datasets/get-query-scale-out-sync-status-in-group>`_.

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    Tuple[pandas.DataFrame, pandas.DataFrame]
        2 pandas dataframes showing the query scale-out sync status.
    """

    columns = {
        "Scale Out Status": "string",
        "Sync Start Time": "datetime",
        "Sync End Time": "datetime",
        "Commit Version": "int",
        "Commit Timestamp": "datetime",
        "Target Sync Version": "int",
        "Target Sync Timestamp": "datetime",
        "Trigger Reason": "string",
        "Min Active Read Version": "int",
        "Min Active Read Timestamp": "datetime",
    }
    df = _create_dataframe(columns=columns)

    columns_rep = {
        "Replica ID": "string",
        "Replica Type": "string",
        "Replica Version": "string",
        "Replica Timestamp": "datetime",
    }

    dfRep = _create_dataframe(columns=columns_rep)

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (dataset_name, dataset_id) = resolve_dataset_name_and_id(dataset, workspace_id)

    response = _base_api(
        request=f"/v1.0/myorg/groups/{workspace_id}/datasets/{dataset_id}/queryScaleOut/syncStatus"
    )

    o = response.json()
    sos = o.get("scaleOutStatus")

    if sos == "Enabled":
        new_data = {
            "Scale Out Status": o.get("scaleOutStatus"),
            "Sync Start Time": o.get("syncStartTime"),
            "Sync End Time": o.get("syncEndTime"),
            "Commit Version": o.get("commitVersion"),
            "Commit Timestamp": o.get("commitTimestamp"),
            "Target Sync Version": o.get("targetSyncVersion"),
            "Target Sync Timestamp": o.get("targetSyncTimestamp"),
            "Trigger Reason": o.get("triggerReason"),
            "Min Active Read Version": o.get("minActiveReadVersion"),
            "Min Active Read Timestamp": o.get("minActiveReadTimestamp"),
        }
        df = pd.concat([df, pd.DataFrame(new_data, index=[0])], ignore_index=True)

        for r in o.get("scaleOutReplicas", []):
            new_data = {
                "Replica ID": r.get("replicaId"),
                "Replica Type": r.get("replicaType"),
                "Replica Version": str(r.get("replicaVersion")),
                "Replica Timestamp": r.get("replicaTimestamp"),
            }
            dfRep = pd.concat(
                [dfRep, pd.DataFrame(new_data, index=[0])], ignore_index=True
            )

        _update_dataframe_datatypes(dataframe=df, column_map=columns)
        _update_dataframe_datatypes(dataframe=dfRep, column_map=columns_rep)

        return df, dfRep
    else:
        print(f"{sos}\n\n")
        return df, dfRep


@log
def disable_qso(
    dataset: str | UUID, workspace: Optional[str | UUID] = None
) -> pd.DataFrame:
    """
    Sets the max read-only replicas to 0, disabling query scale out.

    This is a wrapper function for the following API: `Datasets - Update Dataset In Group <https://learn.microsoft.com/rest/api/power-bi/datasets/update-dataset-in-group>`_.

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the current query scale out settings.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (dataset_name, dataset_id) = resolve_dataset_name_and_id(dataset, workspace_id)

    payload = {"queryScaleOutSettings": {"maxReadOnlyReplicas": "0"}}

    _base_api(
        request=f"/v1.0/myorg/groups/{workspace_id}/datasets/{dataset_id}",
        method="patch",
        payload=payload,
    )

    df = list_qso_settings(dataset=dataset_id, workspace=workspace_id)

    print(
        f"{icons.green_dot} Query scale out has been disabled for the '{dataset_name}' semantic model within the '{workspace_name}' workspace."
    )

    return df


@log
def set_qso(
    dataset: str | UUID,
    auto_sync: bool = True,
    max_read_only_replicas: int = -1,
    workspace: Optional[str | UUID] = None,
) -> pd.DataFrame:
    """
    Sets the query scale out settings for a semantic model.

    This is a wrapper function for the following API: `Datasets - Update Dataset In Group <https://learn.microsoft.com/rest/api/power-bi/datasets/update-dataset-in-group>`_.

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    auto_sync : bool, default=True
        Whether the semantic model automatically syncs read-only replicas.
    max_read_only_replicas : int, default=-1
        To enable semantic model scale-out, set max_read_only_replicas to -1, or any non-0 value. A value of -1 allows Power BI to create as many read-only replicas as your Power BI capacity supports. You can also explicitly set the replica count to a value lower than that of the capacity maximum. Setting max_read_only_replicas to -1 is recommended.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the current query scale-out settings.
    """

    from sempy_labs._helper_functions import is_default_semantic_model

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (dataset_name, dataset_id) = resolve_dataset_name_and_id(dataset, workspace_id)

    if is_default_semantic_model(dataset=dataset_id, workspace=workspace_id):
        raise ValueError(
            f"{icons.red_dot} The 'set_qso' function does not run against default semantic models."
        )

    if max_read_only_replicas == 0:
        disable_qso(dataset=dataset_id, workspace=workspace_id)
        return

    payload = {
        "queryScaleOutSettings": {
            "autoSyncReadOnlyReplicas": auto_sync,
            "maxReadOnlyReplicas": max_read_only_replicas,
        }
    }

    dfL = list_qso_settings(dataset=dataset_id, workspace=workspace_id)
    storage_mode = dfL["Storage Mode"].iloc[0]

    if storage_mode == "Small":
        set_semantic_model_storage_format(
            dataset=dataset_id, storage_format="Large", workspace=workspace_id
        )

    _base_api(
        request=f"/v1.0/myorg/groups/{workspace_id}/datasets/{dataset_id}",
        method="patch",
        payload=payload,
    )

    df = list_qso_settings(dataset=dataset_id, workspace=workspace_id)
    print(
        f"{icons.green_dot} Query scale out has been set on the '{dataset_name}' semantic model within the '{workspace_name}' workspace."
    )

    return df


@log
def set_semantic_model_storage_format(
    dataset: str | UUID, storage_format: str, workspace: Optional[str | UUID] = None
):
    """
    Sets the semantic model storage format.

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    storage_format : str
        The storage format for the semantic model. Valid options: 'Large', 'Small'.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (dataset_name, dataset_id) = resolve_dataset_name_and_id(dataset, workspace_id)

    storage_format = storage_format.capitalize()

    if storage_format == "Abf":
        storage_format = "Small"
    elif storage_format.startswith("Premium"):
        storage_format = "Large"

    storageFormats = ["Small", "Large"]

    if storage_format == "Large":
        payload = {"targetStorageMode": "PremiumFiles"}
    elif storage_format == "Small":
        payload = {"targetStorageMode": "Abf"}
    else:
        raise ValueError(
            f"{icons.red_dot} Invalid storage format value. Valid options: {storageFormats}."
        )

    dfL = list_qso_settings(dataset=dataset_id, workspace=workspace_id)
    current_storage_format = dfL["Storage Mode"].iloc[0]

    if current_storage_format == storage_format:
        print(
            f"{icons.info} The '{dataset_name}' semantic model within the '{workspace_name}' workspace is already set to '{storage_format.lower()}' storage format."
        )
        return

    _base_api(
        request=f"/v1.0/myorg/groups/{workspace_id}/datasets/{dataset_id}",
        method="patch",
        payload=payload,
    )
    print(
        f"{icons.green_dot} The semantic model storage format for the '{dataset_name}' semantic model within the '{workspace_name}' workspace has been set to '{storage_format}'."
    )


@log
def list_qso_settings(
    dataset: Optional[str | UUID] = None, workspace: Optional[str | UUID] = None
) -> pd.DataFrame:
    """
    Shows the query scale out settings for a semantic model (or all semantic models within a workspace).

    Parameters
    ----------
    dataset : str | uuid.UUID, default=None
        Name or ID of the semantic model.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the query scale out settings.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    if dataset is not None:
        (dataset_name, dataset_id) = resolve_dataset_name_and_id(dataset, workspace_id)

    columns = {
        "Dataset Id": "string",
        "Dataset Name": "string",
        "Storage Mode": "string",
        "QSO Auto Sync Enabled": "bool",
        "QSO Max Read Only Replicas": "int",
    }
    df = _create_dataframe(columns=columns)

    response = _base_api(request=f"/v1.0/myorg/groups/{workspace_id}/datasets")

    for v in response.json().get("value", []):
        tsm = v.get("targetStorageMode")
        if tsm == "Abf":
            sm = "Small"
        else:
            sm = "Large"
        new_data = {
            "Dataset Id": v.get("id"),
            "Dataset Name": v.get("name"),
            "Storage Mode": sm,
            "QSO Auto Sync Enabled": v.get("queryScaleOutSettings", {}).get(
                "autoSyncReadOnlyReplicas"
            ),
            "QSO Max Read Only Replicas": v.get("queryScaleOutSettings", {}).get(
                "maxReadOnlyReplicas"
            ),
        }
        df = pd.concat([df, pd.DataFrame(new_data, index=[0])], ignore_index=True)

    _update_dataframe_datatypes(dataframe=df, column_map=columns)

    if dataset is not None:
        df = df[df["Dataset Id"] == dataset_id]

    return df


@log
def set_workspace_default_storage_format(
    storage_format: str, workspace: Optional[str | UUID] = None
):
    """
    Sets the default storage format for semantic models within a workspace.

    Parameters
    ----------
    storage_format : str
        The storage format for the semantic model. Valid options: 'Large', 'Small'.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    # https://learn.microsoft.com/en-us/rest/api/power-bi/groups/update-group#defaultdatasetstorageformat

    storageFormats = ["Small", "Large"]
    storage_format = storage_format.capitalize()

    if storage_format not in storageFormats:
        raise ValueError(
            f"{icons.red_dot} Invalid storage format. Please choose from these options: {storageFormats}."
        )

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    # Check current storage format
    dfW = fabric.list_workspaces(filter=f"name eq '{workspace_name}'")
    if len(dfW) == 0:
        raise ValueError()
    current_storage_format = dfW["Default Dataset Storage Format"].iloc[0]

    if current_storage_format == storage_format:
        print(
            f"{icons.info} The '{workspace_name}' is already set to a default storage format of '{current_storage_format}'."
        )
        return

    payload = {
        "name": workspace_name,
        "defaultDatasetStorageFormat": storage_format,
    }

    _base_api(
        request=f"/v1.0/myorg/groups/{workspace_id}", method="patch", payload=payload
    )

    print(
        f"{icons.green_dot} The default storage format for the '{workspace_name}' workspace has been updated to '{storage_format}."
    )



from sempy_labs.lakehouse._get_lakehouse_columns import (
    get_lakehouse_columns,
)
from sempy_labs.lakehouse._get_lakehouse_tables import (
    get_lakehouse_tables,
)
from sempy_labs.lakehouse._lakehouse import (
    lakehouse_attached,
    optimize_lakehouse_tables,
    vacuum_lakehouse_tables,
    run_table_maintenance,
)
from sempy_labs.lakehouse._shortcuts import (
    # create_shortcut,
    create_shortcut_onelake,
    delete_shortcut,
    reset_shortcut_cache,
    list_shortcuts,
)
from sempy_labs.lakehouse._blobs import (
    recover_lakehouse_object,
    list_blobs,
)

__all__ = [
    "get_lakehouse_columns",
    "get_lakehouse_tables",
    "lakehouse_attached",
    "optimize_lakehouse_tables",
    # create_shortcut,
    "create_shortcut_onelake",
    "delete_shortcut",
    "vacuum_lakehouse_tables",
    "reset_shortcut_cache",
    "run_table_maintenance",
    "list_shortcuts",
    "recover_lakehouse_object",
    "list_blobs",
]



import pandas as pd
from sempy_labs._helper_functions import (
    _base_api,
)
from uuid import UUID
from typing import Optional
from sempy_labs.admin._basic_functions import (
    _resolve_workspace_name_and_id,
    _create_dataframe,
    _update_dataframe_datatypes,
)


def list_unused_artifacts(workspace: Optional[str | UUID] = None) -> pd.DataFrame:
    """
    Returns a list of datasets, reports, and dashboards that have not been used within 30 days for the specified workspace.

    This is a wrapper function for the following API: `Admin - Groups GetUnusedArtifactsAsAdmin <https://learn.microsoft.com/rest/api/power-bi/admin/groups-get-unused-artifacts-as-admin>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of datasets, reports, and dashboards that have not been used within 30 days for the specified workspace.
    """

    (workspace_name, workspace_id) = _resolve_workspace_name_and_id(workspace)

    columns = {
        "Artifact Name": "string",
        "Artifact Id": "string",
        "Artifact Type": "string",
        "Artifact Size in MB": "int",
        "Created Date Time": "datetime",
        "Last Accessed Date Time": "datetime",
    }

    df = _create_dataframe(columns=columns)

    responses = _base_api(
        request=f"/v1.0/myorg/admin/groups/{workspace_id}/unused",
        client="fabric_sp",
        uses_pagination=True,
    )

    for r in responses:
        for i in r.get("unusedArtifactEntities", []):
            new_data = {
                "Artifact Name": i.get("artifactId"),
                "Artifact Id": i.get("displayName"),
                "Artifact Type": i.get("artifactType"),
                "Artifact Size in MB": i.get("artifactSizeInMB"),
                "Created Date Time": i.get("createdDateTime"),
                "Last Accessed Date Time": i.get("lastAccessedDateTime"),
            }

            df = pd.concat([df, pd.DataFrame(new_data, index=[0])], ignore_index=True)

    _update_dataframe_datatypes(dataframe=df, column_map=columns)

    return df



import sempy.fabric as fabric
import pandas as pd
from sempy_labs._helper_functions import (
    format_dax_object_name,
    resolve_workspace_name_and_id,
    resolve_dataset_name_and_id,
    resolve_workspace_name,
)
from IPython.display import display
from sempy_labs.lakehouse import get_lakehouse_columns
from sempy_labs.directlake._dl_helper import get_direct_lake_source
from typing import Optional
import sempy_labs._icons as icons
from sempy._utils._log import log
from uuid import UUID


@log
def direct_lake_schema_compare(
    dataset: str | UUID,
    workspace: Optional[str | UUID] = None,
    **kwargs,
):
    """
    Checks that all the tables in a Direct Lake semantic model map to tables in their corresponding lakehouse and that the columns in each table exist.

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    if "lakehouse" in kwargs:
        print(
            "The 'lakehouse' parameter has been deprecated as it is no longer necessary. Please remove this parameter from the function going forward."
        )
        del kwargs["lakehouse"]
    if "lakehouse_workspace" in kwargs:
        print(
            "The 'lakehouse_workspace' parameter has been deprecated as it is no longer necessary. Please remove this parameter from the function going forward."
        )
        del kwargs["lakehouse_workspace"]

    fabric.refresh_tom_cache(workspace=workspace)

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (dataset_name, dataset_id) = resolve_dataset_name_and_id(dataset, workspace_id)

    artifact_type, lakehouse_name, lakehouse_id, lakehouse_workspace_id = (
        get_direct_lake_source(dataset=dataset_id, workspace=workspace_id)
    )

    if artifact_type == "Warehouse":
        raise ValueError(
            f"{icons.red_dot} This function is only valid for Direct Lake semantic models which source from Fabric lakehouses (not warehouses)."
        )

    dfP = fabric.list_partitions(dataset=dataset_id, workspace=workspace_id)

    if not any(r["Mode"] == "DirectLake" for _, r in dfP.iterrows()):
        raise ValueError(
            f"{icons.red_dot} The '{dataset_name}' semantic model within the '{workspace_name}' workspace is not in Direct Lake mode."
        )

    if artifact_type is None:
        raise ValueError(
            f"{icons.red_dot} This function only supports Direct Lake semantic models where the source lakehouse resides in the same workpace as the semantic model."
        )

    lakehouse_workspace = resolve_workspace_name(workspace_id=lakehouse_workspace_id)
    dfT = fabric.list_tables(dataset=dataset_id, workspace=workspace_id)
    dfC = fabric.list_columns(dataset=dataset_id, workspace=workspace_id)
    lc = get_lakehouse_columns(lakehouse_name, lakehouse_workspace)

    dfT.rename(columns={"Type": "Table Type"}, inplace=True)
    dfP_filt = dfP[dfP["Mode"] == "DirectLake"]
    dfC = pd.merge(dfC, dfP[["Table Name", "Query"]], on="Table Name", how="inner")
    dfC = pd.merge(
        dfC,
        dfT[["Name", "Table Type"]],
        left_on="Table Name",
        right_on="Name",
        how="inner",
    )
    dfC["Full Column Name"] = format_dax_object_name(dfC["Query"], dfC["Source"])
    dfC_filt = dfC[dfC["Table Type"] == "Table"]
    # Schema compare
    missingtbls = dfP_filt[~dfP_filt["Query"].isin(lc["Table Name"])]
    missingtbls = missingtbls[["Table Name", "Query"]]
    missingtbls.rename(columns={"Query": "Source Table"}, inplace=True)
    missingcols = dfC_filt[~dfC_filt["Full Column Name"].isin(lc["Full Column Name"])]
    missingcols = missingcols[
        ["Table Name", "Column Name", "Type", "Data Type", "Source"]
    ]
    missingcols.rename(columns={"Source": "Source Column"}, inplace=True)

    if len(missingtbls) == 0:
        print(
            f"{icons.green_dot} All tables exist in the '{lakehouse_name}' lakehouse within the '{lakehouse_workspace}' workspace."
        )
    else:
        print(
            f"{icons.yellow_dot} The following tables exist in the '{dataset_name}' semantic model within the '{workspace_name}' workspace"
            f" but do not exist in the '{lakehouse_name}' lakehouse within the '{lakehouse_workspace}' workspace."
        )
        display(missingtbls)
    if len(missingcols) == 0:
        print(
            f"{icons.green_dot} All columns exist in the '{lakehouse_name}' lakehouse within the '{lakehouse_workspace}' workspace."
        )
    else:
        print(
            f"{icons.yellow_dot} The following columns exist in the '{dataset_name}' semantic model within the '{workspace_name}' workspace "
            f"but do not exist in the '{lakehouse_name}' lakehouse within the '{lakehouse_workspace}' workspace."
        )
        display(missingcols)



import sempy.fabric as fabric
import pandas as pd
from IPython.display import display, HTML
import zipfile
import os
import shutil
import datetime
import warnings
from sempy_labs._helper_functions import (
    format_dax_object_name,
    resolve_lakehouse_name,
    save_as_delta_table,
    resolve_workspace_capacity,
    _get_column_aggregate,
    resolve_workspace_name_and_id,
    resolve_dataset_name_and_id,
    _create_spark_session,
    resolve_workspace_id,
    resolve_workspace_name,
)
from sempy_labs._list_functions import list_relationships, list_tables
from sempy_labs.lakehouse import lakehouse_attached, get_lakehouse_tables
from sempy_labs.directlake import get_direct_lake_source
from typing import Optional
from sempy._utils._log import log
import sempy_labs._icons as icons
from pathlib import Path
from uuid import UUID


@log
def vertipaq_analyzer(
    dataset: str | UUID,
    workspace: Optional[str | UUID] = None,
    export: Optional[str] = None,
    read_stats_from_data: bool = False,
    **kwargs,
) -> dict[str, pd.DataFrame]:
    """
    Displays an HTML visualization of the `Vertipaq Analyzer <https://www.sqlbi.com/tools/vertipaq-analyzer/>`_ statistics from a semantic model.

    `Vertipaq Analyzer <https://www.sqlbi.com/tools/vertipaq-analyzer/>`_ is an open-sourced tool built by SQLBI. It provides a detailed analysis of the VertiPaq engine, which is the in-memory engine used by Power BI and Analysis Services Tabular models.

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    workspace : str| uuid.UUID, default=None
        The Fabric workspace name or ID in which the semantic model exists.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    export : str, default=None
        Specifying 'zip' will export the results to a zip file in your lakehouse (which can be imported using the import_vertipaq_analyzer function.
        Specifying 'table' will export the results to delta tables (appended) in your lakehouse.
        Default value: None.
    read_stats_from_data : bool, default=False
        Setting this parameter to true has the function get Column Cardinality and Missing Rows using DAX (Direct Lake semantic models achieve this using a Spark query to the lakehouse).

    Returns
    -------
    dict[str, pandas.DataFrame]
        A dictionary of pandas dataframes showing the vertipaq analyzer statistics.
    """

    from sempy_labs.tom import connect_semantic_model

    if "lakehouse_workspace" in kwargs:
        print(
            f"{icons.info} The 'lakehouse_workspace' parameter has been deprecated as it is no longer necessary. Please remove this parameter from the function going forward."
        )
        del kwargs["lakehouse_workspace"]

    pd.options.mode.copy_on_write = True
    warnings.filterwarnings(
        "ignore", message="createDataFrame attempted Arrow optimization*"
    )

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (dataset_name, dataset_id) = resolve_dataset_name_and_id(dataset, workspace_id)

    fabric.refresh_tom_cache(workspace=workspace)

    vertipaq_map = {
        "Model": {
            "Dataset Name": [icons.data_type_string, icons.no_format],
            "Total Size": [icons.data_type_long, icons.int_format],
            "Table Count": [icons.data_type_long, icons.int_format],
            "Column Count": [icons.data_type_long, icons.int_format],
            "Compatibility Level": [icons.data_type_long, icons.no_format],
            "Default Mode": [icons.data_type_string, icons.no_format],
        },
        "Tables": {
            "Table Name": [icons.data_type_string, icons.no_format],
            "Type": [icons.data_type_string, icons.no_format],
            "Row Count": [icons.data_type_long, icons.int_format],
            "Total Size": [icons.data_type_long, icons.int_format],
            "Dictionary Size": [icons.data_type_long, icons.int_format],
            "Data Size": [icons.data_type_long, icons.int_format],
            "Hierarchy Size": [icons.data_type_long, icons.int_format],
            "Relationship Size": [icons.data_type_long, icons.int_format],
            "User Hierarchy Size": [icons.data_type_long, icons.int_format],
            "Partitions": [icons.data_type_long, icons.int_format],
            "Columns": [icons.data_type_long, icons.int_format],
            "% DB": [icons.data_type_double, icons.pct_format],
        },
        "Partitions": {
            "Table Name": [icons.data_type_string, icons.no_format],
            "Partition Name": [icons.data_type_string, icons.no_format],
            "Mode": [icons.data_type_string, icons.no_format],
            "Record Count": [icons.data_type_long, icons.int_format],
            "Segment Count": [icons.data_type_long, icons.int_format],
            "Records per Segment": [icons.data_type_double, icons.int_format],
        },
        "Columns": {
            "Table Name": [icons.data_type_string, icons.no_format],
            "Column Name": [icons.data_type_string, icons.no_format],
            "Type": [icons.data_type_string, icons.no_format],
            "Cardinality": [icons.data_type_long, icons.int_format],
            "Total Size": [icons.data_type_long, icons.int_format],
            "Data Size": [icons.data_type_long, icons.int_format],
            "Dictionary Size": [icons.data_type_long, icons.int_format],
            "Hierarchy Size": [icons.data_type_long, icons.int_format],
            "% Table": [icons.data_type_double, icons.pct_format],
            "% DB": [icons.data_type_double, icons.pct_format],
            "Data Type": [icons.data_type_string, icons.no_format],
            "Encoding": [icons.data_type_string, icons.no_format],
            "Is Resident": [icons.data_type_bool, icons.no_format],
            "Temperature": [icons.data_type_double, icons.int_format],
            "Last Accessed": [icons.data_type_timestamp, icons.no_format],
        },
        "Hierarchies": {
            "Table Name": [icons.data_type_string, icons.no_format],
            "Hierarchy Name": [icons.data_type_string, icons.no_format],
            "Used Size": [icons.data_type_long, icons.int_format],
        },
        "Relationships": {
            "From Object": [icons.data_type_string, icons.no_format],
            "To Object": [icons.data_type_string, icons.no_format],
            "Multiplicity": [icons.data_type_string, icons.no_format],
            "Used Size": [icons.data_type_long, icons.int_format],
            "Max From Cardinality": [icons.data_type_long, icons.int_format],
            "Max To Cardinality": [icons.data_type_long, icons.int_format],
            "Missing Rows": [icons.data_type_long, icons.int_format],
        },
    }

    with connect_semantic_model(
        dataset=dataset_id, workspace=workspace_id, readonly=True
    ) as tom:
        compat_level = tom.model.Model.Database.CompatibilityLevel
        is_direct_lake = tom.is_direct_lake()
        def_mode = tom.model.DefaultMode
        table_count = tom.model.Tables.Count
        column_count = len(list(tom.all_columns()))
        if table_count == 0:
            print(
                f"{icons.warning} The '{dataset_name}' semantic model within the '{workspace_name}' workspace has no tables. Vertipaq Analyzer can only be run if the semantic model has tables."
            )
            return

    dfT = list_tables(dataset=dataset_id, extended=True, workspace=workspace_id)

    dfT.rename(columns={"Name": "Table Name"}, inplace=True)
    columns_to_keep = list(vertipaq_map["Tables"].keys())
    dfT = dfT[dfT.columns.intersection(columns_to_keep)]

    dfC = fabric.list_columns(dataset=dataset_id, extended=True, workspace=workspace_id)
    dfC["Column Object"] = format_dax_object_name(dfC["Table Name"], dfC["Column Name"])
    dfC.rename(columns={"Column Cardinality": "Cardinality"}, inplace=True)
    dfH = fabric.list_hierarchies(
        dataset=dataset_id, extended=True, workspace=workspace_id
    )
    dfR = list_relationships(dataset=dataset_id, extended=True, workspace=workspace_id)
    dfP = fabric.list_partitions(
        dataset=dataset_id, extended=True, workspace=workspace_id
    )

    artifact_type = None
    if is_direct_lake:
        artifact_type, lakehouse_name, lakehouse_id, lakehouse_workspace_id = (
            get_direct_lake_source(dataset=dataset_id, workspace=workspace_id)
        )

    dfR["Missing Rows"] = 0
    dfR["Missing Rows"] = dfR["Missing Rows"].astype(int)

    # Direct Lake
    if read_stats_from_data:
        if is_direct_lake and artifact_type == "Lakehouse":
            dfC = pd.merge(
                dfC,
                dfP[["Table Name", "Query", "Source Type"]],
                on="Table Name",
                how="left",
            )
            dfC_flt = dfC[
                (dfC["Source Type"] == "Entity")
                & (~dfC["Column Name"].str.startswith("RowNumber-"))
            ]

            object_workspace = resolve_workspace_name(
                workspace_id=lakehouse_workspace_id
            )
            current_workspace_id = resolve_workspace_id()
            if current_workspace_id != lakehouse_workspace_id:
                lakeTables = get_lakehouse_tables(
                    lakehouse=lakehouse_name, workspace=object_workspace
                )

            sql_statements = []
            spark = _create_spark_session()
            # Loop through tables
            for lakeTName in dfC_flt["Query"].unique():
                query = "SELECT "
                columns_in_table = dfC_flt.loc[
                    dfC_flt["Query"] == lakeTName, "Source"
                ].unique()

                # Loop through columns within those tables
                for scName in columns_in_table:
                    query = query + f"COUNT(DISTINCT(`{scName}`)) AS `{scName}`, "

                query = query[:-2]
                if lakehouse_workspace_id == current_workspace_id:
                    query = query + f" FROM {lakehouse_name}.{lakeTName}"
                else:
                    lakeTables_filt = lakeTables[lakeTables["Table Name"] == lakeTName]
                    tPath = lakeTables_filt["Location"].iloc[0]

                    df = spark.read.format("delta").load(tPath)
                    tempTableName = "delta_table_" + lakeTName
                    df.createOrReplaceTempView(tempTableName)
                    query = query + f" FROM {tempTableName}"
                sql_statements.append((lakeTName, query))

            for o in sql_statements:
                tName = o[0]
                query = o[1]

                df = spark.sql(query)

                for column in df.columns:
                    x = df.collect()[0][column]
                    for i, r in dfC.iterrows():
                        if r["Query"] == tName and r["Source"] == column:
                            dfC.at[i, "Cardinality"] = x

            # Remove column added temporarily
            dfC.drop(columns=["Query", "Source Type"], inplace=True)

            # Direct Lake missing rows
            dfR = pd.merge(
                dfR,
                dfP[["Table Name", "Query"]],
                left_on="From Table",
                right_on="Table Name",
                how="left",
            )
            dfR.rename(columns={"Query": "From Lake Table"}, inplace=True)
            dfR.drop(columns=["Table Name"], inplace=True)
            dfR = pd.merge(
                dfR,
                dfP[["Table Name", "Query"]],
                left_on="To Table",
                right_on="Table Name",
                how="left",
            )
            dfR.rename(columns={"Query": "To Lake Table"}, inplace=True)
            dfR.drop(columns=["Table Name"], inplace=True)
            dfR = pd.merge(
                dfR,
                dfC[["Column Object", "Source"]],
                left_on="From Object",
                right_on="Column Object",
                how="left",
            )
            dfR.rename(columns={"Source": "From Lake Column"}, inplace=True)
            dfR.drop(columns=["Column Object"], inplace=True)
            dfR = pd.merge(
                dfR,
                dfC[["Column Object", "Source"]],
                left_on="To Object",
                right_on="Column Object",
                how="left",
            )
            dfR.rename(columns={"Source": "To Lake Column"}, inplace=True)
            dfR.drop(columns=["Column Object"], inplace=True)

            spark = _create_spark_session()
            for i, r in dfR.iterrows():
                fromTable = r["From Lake Table"]
                fromColumn = r["From Lake Column"]
                toTable = r["To Lake Table"]
                toColumn = r["To Lake Column"]

                if lakehouse_workspace_id == current_workspace_id:
                    query = f"select count(f.{fromColumn}) as {fromColumn}\nfrom {fromTable} as f\nleft join {toTable} as c on f.{fromColumn} = c.{toColumn}\nwhere c.{toColumn} is null"
                else:
                    tempTableFrom = f"delta_table_{fromTable}"
                    tempTableTo = f"delta_table_{toTable}"

                    query = f"select count(f.{fromColumn}) as {fromColumn}\nfrom {tempTableFrom} as f\nleft join {tempTableTo} as c on f.{fromColumn} = c.{toColumn}\nwhere c.{toColumn} is null"

                # query = f"select count(f.{fromColumn}) as {fromColumn}\nfrom {fromTable} as f\nleft join {toTable} as c on f.{fromColumn} = c.{toColumn}\nwhere c.{toColumn} is null"

                df = spark.sql(query)
                missingRows = df.collect()[0][0]
                dfR.at[i, "Missing Rows"] = missingRows

            dfR["Missing Rows"] = dfR["Missing Rows"].astype(int)
        elif not is_direct_lake:
            # Calculate missing rows using DAX for non-direct lake
            for i, r in dfR.iterrows():
                fromTable = r["From Table"]
                fromColumn = r["From Column"]
                toTable = r["To Table"]
                toColumn = r["To Column"]
                isActive = bool(r["Active"])
                fromObject = format_dax_object_name(fromTable, fromColumn)
                toObject = format_dax_object_name(toTable, toColumn)

                missingRows = 0

                query = f"evaluate\nsummarizecolumns(\n\"1\",calculate(countrows('{fromTable}'),isblank({toObject}))\n)"

                if not isActive:
                    query = f"evaluate\nsummarizecolumns(\n\"1\",calculate(countrows('{fromTable}'),userelationship({fromObject},{toObject}),isblank({toObject}))\n)"

                result = fabric.evaluate_dax(
                    dataset=dataset_id, dax_string=query, workspace=workspace_id
                )

                try:
                    missingRows = result.iloc[0, 0]
                except Exception:
                    pass

                dfR.at[i, "Missing Rows"] = missingRows
            dfR["Missing Rows"] = dfR["Missing Rows"].astype(int)

    table_totals = dfC.groupby("Table Name")["Total Size"].transform("sum")
    db_total_size = dfC["Total Size"].sum()
    dfC["% Table"] = round((dfC["Total Size"] / table_totals) * 100, 2)
    dfC["% DB"] = round((dfC["Total Size"] / db_total_size) * 100, 2)
    columnList = list(vertipaq_map["Columns"].keys())

    dfC = dfC[dfC["Type"] != "RowNumber"].reset_index(drop=True)

    colSize = dfC[columnList].sort_values(by="Total Size", ascending=False)
    temp = dfC[columnList].sort_values(by="Temperature", ascending=False)
    colSize.reset_index(drop=True, inplace=True)
    temp.reset_index(drop=True, inplace=True)

    export_Col = colSize.copy()
    export_Table = dfT.copy()

    #  Relationships
    dfR = pd.merge(
        dfR,
        dfC[["Column Object", "Cardinality"]],
        left_on="From Object",
        right_on="Column Object",
        how="left",
    )
    dfR.rename(columns={"Cardinality": "Max From Cardinality"}, inplace=True)
    dfR = pd.merge(
        dfR,
        dfC[["Column Object", "Cardinality"]],
        left_on="To Object",
        right_on="Column Object",
        how="left",
    )
    dfR.rename(columns={"Cardinality": "Max To Cardinality"}, inplace=True)
    dfR = dfR[
        [
            "From Object",
            "To Object",
            "Multiplicity",
            "Used Size",
            "Max From Cardinality",
            "Max To Cardinality",
            "Missing Rows",
        ]
    ].sort_values(by="Used Size", ascending=False)
    dfR.reset_index(drop=True, inplace=True)
    export_Rel = dfR.copy()

    # Partitions
    dfP = dfP[
        [
            "Table Name",
            "Partition Name",
            "Mode",
            "Record Count",
            "Segment Count",
            # "Records per Segment",
        ]
    ].sort_values(by="Record Count", ascending=False)
    dfP["Records per Segment"] = round(
        dfP["Record Count"] / dfP["Segment Count"], 2
    )  # Remove after records per segment is fixed
    dfP.reset_index(drop=True, inplace=True)
    export_Part = dfP.copy()

    # Hierarchies
    dfH_filt = dfH[dfH["Level Ordinal"] == 0]
    dfH_filt = dfH_filt[["Table Name", "Hierarchy Name", "Used Size"]].sort_values(
        by="Used Size", ascending=False
    )
    dfH_filt.reset_index(drop=True, inplace=True)
    dfH_filt.fillna({"Used Size": 0}, inplace=True)
    dfH_filt["Used Size"] = dfH_filt["Used Size"].astype(int)
    export_Hier = dfH_filt.copy()

    # Model
    # Converting to KB/MB/GB necessitates division by 1024 * 1000.
    if db_total_size >= 1000000000:
        y = db_total_size / (1024**3) * 1000000000
    elif db_total_size >= 1000000:
        y = db_total_size / (1024**2) * 1000000
    elif db_total_size >= 1000:
        y = db_total_size / (1024) * 1000
    else:
        y = db_total_size
    y = round(y)

    dfModel = pd.DataFrame(
        {
            "Dataset Name": dataset_name,
            "Total Size": y,
            "Table Count": table_count,
            "Column Count": column_count,
            "Compatibility Level": compat_level,
            "Default Mode": def_mode,
        },
        index=[0],
    )
    dfModel.reset_index(drop=True, inplace=True)
    dfModel["Default Mode"] = dfModel["Default Mode"].astype(str)
    export_Model = dfModel.copy()

    def _style_columns_based_on_types(dataframe: pd.DataFrame, column_type_mapping):
        # Define formatting functions based on the type mappings
        format_funcs = {
            "int": lambda x: "{:,}".format(x) if pd.notnull(x) else "",
            "pct": lambda x: "{:.2f}%".format(x) if pd.notnull(x) else "",
            "": lambda x: "{}".format(x),
        }

        # Apply the formatting function to each column based on its specified type
        for col, dt in column_type_mapping.items():
            if dt in format_funcs:
                dataframe[col] = dataframe[col].map(format_funcs[dt])

        return dataframe

    dfModel = _style_columns_based_on_types(
        dfModel,
        column_type_mapping={
            key: values[1] for key, values in vertipaq_map["Model"].items()
        },
    )
    dfT = _style_columns_based_on_types(
        dfT,
        column_type_mapping={
            key: values[1] for key, values in vertipaq_map["Tables"].items()
        },
    )
    dfP = _style_columns_based_on_types(
        dfP,
        column_type_mapping={
            key: values[1] for key, values in vertipaq_map["Partitions"].items()
        },
    )
    colSize = _style_columns_based_on_types(
        colSize,
        column_type_mapping={
            key: values[1] for key, values in vertipaq_map["Columns"].items()
        },
    )
    temp = _style_columns_based_on_types(
        temp,
        column_type_mapping={
            key: values[1] for key, values in vertipaq_map["Columns"].items()
        },
    )
    dfR = _style_columns_based_on_types(
        dfR,
        column_type_mapping={
            key: values[1] for key, values in vertipaq_map["Relationships"].items()
        },
    )
    dfH_filt = _style_columns_based_on_types(
        dfH_filt,
        column_type_mapping={
            key: values[1] for key, values in vertipaq_map["Hierarchies"].items()
        },
    )

    dataFrames = {
        "dfModel": dfModel,
        "dfT": dfT,
        "dfP": dfP,
        "colSize": colSize,
        "temp": temp,
        "dfR": dfR,
        "dfH_filt": dfH_filt,
    }

    dfs = {}
    for fileName, df in dataFrames.items():
        dfs[fileName] = df

    if export is None:
        visualize_vertipaq(dfs)
        return {
            "Model Summary": export_Model,
            "Tables": export_Table,
            "Partitions": export_Part,
            "Columns": export_Col,
            "Relationships": export_Rel,
            "Hierarchies": export_Hier,
        }

    # Export vertipaq to delta tables in lakehouse
    if export in ["table", "zip"]:
        if not lakehouse_attached():
            raise ValueError(
                f"{icons.red_dot} In order to save the Vertipaq Analyzer results, a lakehouse must be attached to the notebook. Please attach a lakehouse to this notebook."
            )

    if export == "table":
        lakeTName = "vertipaqanalyzer_model"

        lakeT = get_lakehouse_tables()
        lakeT_filt = lakeT[lakeT["Table Name"] == lakeTName]

        if len(lakeT_filt) == 0:
            runId = 1
        else:
            max_run_id = _get_column_aggregate(table_name=lakeTName)
            runId = max_run_id + 1

        dfMap = {
            "Columns": ["Columns", export_Col],
            "Tables": ["Tables", export_Table],
            "Partitions": ["Partitions", export_Part],
            "Relationships": ["Relationships", export_Rel],
            "Hierarchies": ["Hierarchies", export_Hier],
            "Model": ["Model", export_Model],
        }

        print(
            f"{icons.in_progress} Saving Vertipaq Analyzer to delta tables in the lakehouse...\n"
        )
        now = datetime.datetime.now()
        dfD = fabric.list_datasets(workspace=workspace_id, mode="rest")
        dfD_filt = dfD[dfD["Dataset Id"] == dataset_id]
        configured_by = dfD_filt["Configured By"].iloc[0]
        capacity_id, capacity_name = resolve_workspace_capacity(workspace=workspace_id)

        for key_name, (obj, df) in dfMap.items():
            df["Capacity Name"] = capacity_name
            df["Capacity Id"] = capacity_id
            df["Configured By"] = configured_by
            df["Workspace Name"] = workspace_name
            df["Workspace Id"] = workspace_id
            df["Dataset Name"] = dataset_name
            df["Dataset Id"] = dataset_id
            df["RunId"] = runId
            df["Timestamp"] = now

            colName = "Capacity Name"
            df.insert(0, colName, df.pop(colName))
            colName = "Capacity Id"
            df.insert(1, colName, df.pop(colName))
            colName = "Workspace Name"
            df.insert(2, colName, df.pop(colName))
            colName = "Workspace Id"
            df.insert(3, colName, df.pop(colName))
            colName = "Dataset Name"
            df.insert(4, colName, df.pop(colName))
            colName = "Dataset Id"
            df.insert(5, colName, df.pop(colName))
            colName = "Configured By"
            df.insert(6, colName, df.pop(colName))

            df.columns = df.columns.str.replace(" ", "_")

            schema = {
                "Capacity_Name": icons.data_type_string,
                "Capacity_Id": icons.data_type_string,
                "Workspace_Name": icons.data_type_string,
                "Workspace_Id": icons.data_type_string,
                "Dataset_Name": icons.data_type_string,
                "Dataset_Id": icons.data_type_string,
                "Configured_By": icons.data_type_string,
            }

            schema.update(
                {
                    key.replace(" ", "_"): value[0]
                    for key, value in vertipaq_map[key_name].items()
                }
            )
            schema["RunId"] = icons.data_type_long
            schema["Timestamp"] = icons.data_type_timestamp

            delta_table_name = f"VertipaqAnalyzer_{obj}".lower()
            save_as_delta_table(
                dataframe=df,
                delta_table_name=delta_table_name,
                write_mode="append",
                schema=schema,
                merge_schema=True,
            )

    # Export vertipaq to zip file within the lakehouse
    if export == "zip":
        dataFrames = {
            "dfModel": dfModel,
            "dfT": dfT,
            "dfP": dfP,
            "colSize": colSize,
            "temp": temp,
            "dfR": dfR,
            "dfH_filt": dfH_filt,
        }

        zipFileName = f"{workspace_name}.{dataset_name}.zip"

        folderPath = "/lakehouse/default/Files"
        subFolderPath = os.path.join(folderPath, "VertipaqAnalyzer")
        ext = ".csv"
        if not os.path.exists(subFolderPath):
            os.makedirs(subFolderPath, exist_ok=True)
        zipFilePath = os.path.join(subFolderPath, zipFileName)

        # Create CSV files based on dataframes
        for fileName, df in dataFrames.items():
            filePath = os.path.join(subFolderPath, f"{fileName}{ext}")
            df.to_csv(filePath, index=False)

        # Create a zip file and add CSV files to it
        with zipfile.ZipFile(zipFilePath, "w") as zipf:
            for fileName in dataFrames:
                filePath = os.path.join(subFolderPath, f"{fileName}{ext}")
                zipf.write(filePath, os.path.basename(filePath))

        # Clean up: remove the individual CSV files
        for fileName, df in dataFrames.items():
            filePath = os.path.join(subFolderPath, fileName) + ext
            if os.path.exists(filePath):
                os.remove(filePath)
        print(
            f"{icons.green_dot} The Vertipaq Analyzer info for the '{dataset_name}' semantic model in the '{workspace_name}' workspace has been saved "
            f"to the 'Vertipaq Analyzer/{zipFileName}' in the default lakehouse attached to this notebook."
        )


def visualize_vertipaq(dataframes):

    # Tooltips for columns within the visual
    data = [
        {
            "ViewName": "Model",
            "ColumnName": "Dataset Name",
            "Tooltip": "The name of the semantic model",
        },
        {
            "ViewName": "Model",
            "ColumnName": "Total Size",
            "Tooltip": "The size of the model (in bytes)",
        },
        {
            "ViewName": "Model",
            "ColumnName": "Table Count",
            "Tooltip": "The number of tables in the semantic model",
        },
        {
            "ViewName": "Model",
            "ColumnName": "Column Count",
            "Tooltip": "The number of columns in the semantic model",
        },
        {
            "ViewName": "Model",
            "ColumnName": "Compatibility Level",
            "Tooltip": "The compatibility level of the semantic model",
        },
        {
            "ViewName": "Model",
            "ColumnName": "Default Mode",
            "Tooltip": "The default query mode of the semantic model",
        },
        {
            "ViewName": "Table",
            "ColumnName": "Table Name",
            "Tooltip": "The name of the table",
        },
        {"ViewName": "Table", "ColumnName": "Type", "Tooltip": "The type of table"},
        {
            "ViewName": "Table",
            "ColumnName": "Row Count",
            "Tooltip": "The number of rows in the table",
        },
        {
            "ViewName": "Table",
            "ColumnName": "Total Size",
            "Tooltip": "Data Size + Dictionary Size + Hierarchy Size (in bytes)",
        },
        {
            "ViewName": "Table",
            "ColumnName": "Data Size",
            "Tooltip": "The size of the data for all the columns in this table (in bytes)",
        },
        {
            "ViewName": "Table",
            "ColumnName": "Dictionary Size",
            "Tooltip": "The size of the column's dictionary for all columns in this table (in bytes)",
        },
        {
            "ViewName": "Table",
            "ColumnName": "Hierarchy Size",
            "Tooltip": "The size of hierarchy structures for all columns in this table (in bytes)",
        },
        {
            "ViewName": "Table",
            "ColumnName": "% DB",
            "Tooltip": "The size of the table relative to the size of the semantic model",
        },
        {
            "ViewName": "Table",
            "ColumnName": "Partitions",
            "Tooltip": "The number of partitions in the table",
        },
        {
            "ViewName": "Table",
            "ColumnName": "Columns",
            "Tooltip": "The number of columns in the table",
        },
        {
            "ViewName": "Partition",
            "ColumnName": "Table Name",
            "Tooltip": "The name of the table",
        },
        {
            "ViewName": "Partition",
            "ColumnName": "Partition Name",
            "Tooltip": "The name of the partition within the table",
        },
        {
            "ViewName": "Partition",
            "ColumnName": "Mode",
            "Tooltip": "The query mode of the partition",
        },
        {
            "ViewName": "Partition",
            "ColumnName": "Record Count",
            "Tooltip": "The number of rows in the partition",
        },
        {
            "ViewName": "Partition",
            "ColumnName": "Segment Count",
            "Tooltip": "The number of segments within the partition",
        },
        {
            "ViewName": "Partition",
            "ColumnName": "Records per Segment",
            "Tooltip": "The number of rows per segment",
        },
        {
            "ViewName": "Column",
            "ColumnName": "Table Name",
            "Tooltip": "The name of the table",
        },
        {
            "ViewName": "Column",
            "ColumnName": "Column Name",
            "Tooltip": "The name of the column",
        },
        {
            "ViewName": "Column",
            "ColumnName": "Type",
            "Tooltip": "The type of column",
        },
        {
            "ViewName": "Column",
            "ColumnName": "Cardinality",
            "Tooltip": "The number of unique rows in the column",
        },
        {
            "ViewName": "Column",
            "ColumnName": "Total Size",
            "Tooltip": "Data Size + Dictionary Size + Hierarchy Size (in bytes)",
        },
        {
            "ViewName": "Column",
            "ColumnName": "Data Size",
            "Tooltip": "The size of the data for the column (in bytes)",
        },
        {
            "ViewName": "Column",
            "ColumnName": "Dictionary Size",
            "Tooltip": "The size of the column's dictionary (in bytes)",
        },
        {
            "ViewName": "Column",
            "ColumnName": "Hierarchy Size",
            "Tooltip": "The size of hierarchy structures (in bytes)",
        },
        {
            "ViewName": "Column",
            "ColumnName": "% Table",
            "Tooltip": "The size of the column relative to the size of the table",
        },
        {
            "ViewName": "Column",
            "ColumnName": "% DB",
            "Tooltip": "The size of the column relative to the size of the semantic model",
        },
        {
            "ViewName": "Column",
            "ColumnName": "Data Type",
            "Tooltip": "The data type of the column",
        },
        {
            "ViewName": "Column",
            "ColumnName": "Encoding",
            "Tooltip": "The encoding type for the column",
        },
        {
            "ViewName": "Column",
            "ColumnName": "Is Resident",
            "Tooltip": "Indicates whether the column is in memory or not",
        },
        {
            "ViewName": "Column",
            "ColumnName": "Temperature",
            "Tooltip": "A decimal indicating the frequency and recency of queries against the column",
        },
        {
            "ViewName": "Column",
            "ColumnName": "Last Accessed",
            "Tooltip": "The time the column was last queried",
        },
        {
            "ViewName": "Hierarchy",
            "ColumnName": "Table Name",
            "Tooltip": "The name of the table",
        },
        {
            "ViewName": "Hierarchy",
            "ColumnName": "Hierarchy Name",
            "Tooltip": "The name of the hierarchy",
        },
        {
            "ViewName": "Hierarchy",
            "ColumnName": "Used Size",
            "Tooltip": "The size of user hierarchy structures (in bytes)",
        },
        {
            "ViewName": "Relationship",
            "ColumnName": "From Object",
            "Tooltip": "The from table/column in the relationship",
        },
        {
            "ViewName": "Relationship",
            "ColumnName": "To Object",
            "Tooltip": "The to table/column in the relationship",
        },
        {
            "ViewName": "Relationship",
            "ColumnName": "Multiplicity",
            "Tooltip": "The cardinality on each side of the relationship",
        },
        {
            "ViewName": "Relationship",
            "ColumnName": "Used Size",
            "Tooltip": "The size of the relationship (in bytes)",
        },
        {
            "ViewName": "Relationship",
            "ColumnName": "Max From Cardinality",
            "Tooltip": "The number of unique values in the column used in the from side of the relationship",
        },
        {
            "ViewName": "Relationship",
            "ColumnName": "Max To Cardinality",
            "Tooltip": "The number of unique values in the column used in the to side of the relationship",
        },
        {
            "ViewName": "Relationship",
            "ColumnName": "Missing Rows",
            "Tooltip": "The number of rows in the 'from' table which do not map to the key column in the 'to' table",
        },
    ]

    # Create DataFrame
    tooltipDF = pd.DataFrame(data)

    # define the dictionary with {"Tab name":df}
    df_dict = {
        "Model Summary": dataframes["dfModel"],
        "Tables": dataframes["dfT"],
        "Partitions": dataframes["dfP"],
        "Columns (Total Size)": dataframes["colSize"],
        "Columns (Temperature)": dataframes["temp"],
        "Relationships": dataframes["dfR"],
        "Hierarchies": dataframes["dfH_filt"],
    }

    mapping = {
        "Model Summary": "Model",
        "Tables": "Table",
        "Partitions": "Partition",
        "Columns (Total Size)": "Column",
        "Columns (Temperature)": "Column",
        "Relationships": "Relationship",
        "Hierarchies": "Hierarchy",
    }

    # Basic styles for the tabs and tab content
    styles = """
    <style>
        .tab { overflow: hidden; border: 1px solid #ccc; background-color: #f1f1f1; }
        .tab button { background-color: inherit; float: left; border: none; outline: none; cursor: pointer; padding: 14px 16px; transition: 0.3s; }
        .tab button:hover { background-color: #ddd; }
        .tab button.active { background-color: #ccc; }
        .tabcontent { display: none; padding: 6px 12px; border: 1px solid #ccc; border-top: none; }
    </style>
    """
    # JavaScript for tab functionality
    script = """
    <script>
    function openTab(evt, tabName) {
        var i, tabcontent, tablinks;
        tabcontent = document.getElementsByClassName("tabcontent");
        for (i = 0; i < tabcontent.length; i++) {
            tabcontent[i].style.display = "none";
        }
        tablinks = document.getElementsByClassName("tablinks");
        for (i = 0; i < tablinks.length; i++) {
            tablinks[i].className = tablinks[i].className.replace(" active", "");
        }
        document.getElementById(tabName).style.display = "block";
        evt.currentTarget.className += " active";
    }
    </script>
    """

    # HTML for tabs
    tab_html = '<div class="tab">'
    content_html = ""
    for i, (title, df) in enumerate(df_dict.items()):
        tab_id = f"tab{i}"
        tab_html += f'<button class="tablinks" onclick="openTab(event, \'{tab_id}\')">{title}</button>'

        vw = mapping.get(title)

        df_html = df.to_html()
        for col in df.columns:
            tt = None
            try:
                tooltipDF_filt = tooltipDF[
                    (tooltipDF["ViewName"] == vw) & (tooltipDF["ColumnName"] == col)
                ]
                tt = tooltipDF_filt["Tooltip"].iloc[0]
            except Exception:
                pass
            df_html = df_html.replace(f"<th>{col}</th>", f'<th title="{tt}">{col}</th>')
        content_html += (
            f'<div id="{tab_id}" class="tabcontent"><h3>{title}</h3>{df_html}</div>'
        )
    tab_html += "</div>"

    # Display the tabs, tab contents, and run the script
    display(HTML(styles + tab_html + content_html + script))
    # Default to open the first tab
    display(
        HTML("<script>document.getElementsByClassName('tablinks')[0].click();</script>")
    )


@log
def import_vertipaq_analyzer(folder_path: str, file_name: str):
    """
    Imports and visualizes the vertipaq analyzer info from a saved .zip file in your lakehouse.

    Parameters
    ----------
    folder_path : str
        The folder within your lakehouse in which the .zip file containing the vertipaq analyzer info has been saved.
    file_name : str
        The file name of the file which contains the vertipaq analyzer info.

    Returns
    -------
    str
       A visualization of the Vertipaq Analyzer statistics.
    """

    pd.options.mode.copy_on_write = True

    zipFilePath = os.path.join(folder_path, file_name)
    extracted_dir = os.path.join(folder_path, "extracted_dataframes")

    with zipfile.ZipFile(zipFilePath, "r") as zip_ref:
        zip_ref.extractall(extracted_dir)

    # Read all CSV files into a dictionary of DataFrames
    dfs = {}
    for file_name in zip_ref.namelist():
        df = pd.read_csv(extracted_dir + "/" + file_name)
        file_path = Path(file_name)
        df_name = file_path.stem
        dfs[df_name] = df

    visualize_vertipaq(dfs)

    # Clean up: remove the extracted directory
    shutil.rmtree(extracted_dir)



import pandas as pd
from uuid import UUID
from typing import Optional
from sempy_labs._helper_functions import (
    _base_api,
    _create_dataframe,
    resolve_workspace_name_and_id,
    create_item,
)


def list_graphql_apis(workspace: Optional[str | UUID]) -> pd.DataFrame:
    """
    Shows the Graph QL APIs within a workspace.

    This is a wrapper function for the following API: `Items - List GraphQLApis <https://learn.microsoft.com/rest/api/fabric/graphqlapi/items/list-graphqlapi-s>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the GraphQL APIs within a workspace.
    """

    columns = {
        "GraphQL API Name": "string",
        "GraphQL API Id": "string",
        "Description": "string",
    }
    df = _create_dataframe(columns=columns)

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    responses = _base_api(
        request=f"/v1/workspaces/{workspace_id}/GraphQLApis",
        uses_pagination=True,
        client="fabric_sp",
    )

    for r in responses:
        for v in r.get("value", []):
            new_data = {
                "GraphQL API Name": v.get("displayName"),
                "GraphQL API Id": v.get("id"),
                "Description": v.get("description"),
            }
            df = pd.concat([df, pd.DataFrame(new_data, index=[0])], ignore_index=True)

    return df


def create_graphql_api(
    name: str, description: Optional[str] = None, workspace: Optional[str | UUID] = None
):
    """
    Creates a GraphQL API.

    This is a wrapper function for the following API: `Items - Create GraphQLApi <https://learn.microsoft.com/rest/api/fabric/graphqlapi/items/create-graphqlapi>`_.

    Parameters
    ----------
    name: str
        Name of the GraphQL API.
    description : str, default=None
        A description of the GraphQL API.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    create_item(
        name=name, description=description, type="GraphQLApi", workspace=workspace
    )



import sempy
from uuid import UUID
import sempy_labs._icons as icons
from typing import Optional


def migrate_direct_lake_to_import(
    dataset: str | UUID,
    workspace: Optional[str | UUID] = None,
    mode: str = "import",
):
    """
    Migrates a semantic model or specific table(s) from a Direct Lake mode to import or DirectQuery mode. After running this function, you must go to the semantic model settings and update the cloud connection. Not doing so will result in an inablity to refresh/use the semantic model.

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    mode : str, default="import"
        The mode to migrate to. Can be either "import" or "directquery".
    """

    sempy.fabric._client._utils._init_analysis_services()
    import Microsoft.AnalysisServices.Tabular as TOM
    from sempy_labs.tom import connect_semantic_model

    modes = {
        "import": "Import",
        "directquery": "DirectQuery",
        "dq": "DirectQuery",
    }

    # Resolve mode
    mode = mode.lower()
    actual_mode = modes.get(mode)
    if actual_mode is None:
        raise ValueError(f"Invalid mode '{mode}'. Must be one of {list(modes.keys())}.")

    # if isinstance(tables, str):
    #     tables = [tables]

    with connect_semantic_model(
        dataset=dataset, workspace=workspace, readonly=False
    ) as tom:

        if not tom.is_direct_lake():
            print(
                f"{icons.warning} The '{dataset}' semantic model within the '{workspace}' workspace is not in Direct Lake mode."
            )
            return

        # if tables is None:
        table_list = [t for t in tom.model.Tables]
        # else:
        #    table_list = [t for t in tom.model.Tables if t.Name in tables]
        # if not table_list:
        #    raise ValueError(f"{icons.red_dot} No tables found to migrate.")

        for t in table_list:
            table_name = t.Name
            if t.Partitions.Count == 1 and all(
                p.Mode == TOM.ModeType.DirectLake for p in t.Partitions
            ):
                p = next(p for p in t.Partitions)
                partition_name = p.Name
                entity_name = p.Source.EntityName
                schema_name = p.Source.SchemaName or "dbo"
                # Rename Direct Lake partition
                t.Partitions[partition_name].Name = f"{partition_name}_remove"

                # Generate M expression for import partition
                expression = f"""let\n\tSource = DatabaseQuery,\n\tData = Source{{[Schema="{schema_name}",Item="{entity_name}"]}}[Data]\nin\n\tData"""

                # Generate M partition
                tom.add_m_partition(
                    table_name=table_name,
                    partition_name=partition_name,
                    expression=expression,
                    mode=actual_mode,
                )
                # Remove Direct Lake partition
                tom.remove_object(object=p)
                # if tables is not None:
                #    print(
                #        f"{icons.green_dot} The '{table_name}' table has been migrated to '{actual_mode}' mode."
                #    )

        tom.model.Model.DefaultMode = TOM.ModeType.Import
    # if tables is None:
    print(
        f"{icons.green_dot} All tables which were in Direct Lake mode have been migrated to '{actual_mode}' mode."
    )

    # Check
    # for t in tom.model.Tables:
    #    if t.Partitions.Count == 1 and all(p.Mode == TOM.ModeType.Import for p in t.Partitions) and t.CalculationGroup is None:
    #        p = next(p for p in t.Partitions)
    #        print(p.Name)
    #        print(p.Source.Expression)
